import sys
import os
import os.path
import re
import textwrap
import copy


try:
    from .__version__ import __version__, __version_info__, __interface__
except ModuleNotFoundError:
    # in an out-of-tree build, __version__.py will be generated in the build dir
    from __version__ import __version__, __version_info__, __interface__

__all__ = [
    'pluralize',
    'doxygenparams',
    'Type',
    'RefType',
    'CompressedType',
    'EnumEntry',
    'Enum',
    'GeneralEnum',
    'DefEnum',
    'EventEnum',
    'Attribute',
    'ArrayAttribute',
    'DefRefAttribute',
    'SelfRefAttribute',
    'EnumAttribute',
    'Record',
    'DefRecord',
    'LocalOnlyDefRecord',
    'GlobalOnlyDefRecord',
    'EventRecord',
    'Version',
    '__version__',
    '__version_info__',
    '__interface__',
    'types',
    'enums',
    'defs',
    'events',
    'snaps',
    'attic_enums'
]

def _execfile(filepath, globals=globals(), locals=None):
    if globals is None:
        globals = {}
    globals.update({
        "__file__": filepath,
        "__name__": "__main__",
    })
    with open(filepath, 'rb') as file:
        exec(compile(file.read(), filepath, 'exec'), globals, locals)

def _doxygendoc(doc, prefix, width):
    pars = []
    for par in re.split('\\n{2,}', doc):
        if not par:
            continue
        if par[:10] == '<no-wrap>\n' and par[-11:] == '\n</no-wrap>':
            par = ('\n' + prefix).join(re.split('\n', par[10:-11]))
        else:
            par = textwrap.fill(par,
                    width=width,
                    initial_indent='',
                    subsequent_indent=prefix,
                    break_long_words=False)
        pars.append(par)
    return ('\n' + prefix + '\n' + prefix).join(pars)

def pluralize(name):
    if name[-1] == 's' and not name[-2] == 's':
        return name
    if name[-2:] == 'ss':
        return name + 'es'
    if name[-1] == 'y':
        return name[:-1] + 'ies'
    else:
        return name + 's'

def doxygenparams(parameters,
                  prefix=' *  ',
                  initial_prefix='\n *  '):
    """Generates doxygen-style @param list for the (parameter,doc) tuples in parameters."""

    max_attr_len = 0
    for parameter, doc in parameters:
        if len(parameter) > max_attr_len:
            max_attr_len = len(parameter)

    if len(parameters) == 0:
        return ''

    result = []
    for parameter, doc in parameters:
        result.append(
            textwrap.fill(
                doc,
                width=77,
                initial_indent='@param %-*s ' % (max_attr_len, parameter),
                subsequent_indent='%s%s' % (prefix, ' ' * (8 + max_attr_len)),
                break_long_words=False))
    return initial_prefix + ('\n%s' % (prefix,)).join(result)

class streqobject(object):
    """Object which can be compared against a string."""
    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        else:
            return id(self) == id(other)
    def __hash__(self):
        return str(self).__hash__()


class klist(list):
    """List where the entries can also be accessed by string."""
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)
        self._n2i = {}

    def __getitem__(self, key):
        if isinstance(key, str):
            return self[self.index(key)]
        else:
            return list.__getitem__(self, key)

types = klist()
enums = klist()
defs = klist()
events = klist()
snaps = klist()
attic_enums = klist()

_camel_case_name_re = re.compile(r'([A-Z]+[a-z0-9]*)*$')
_lower_camel_case_name_re = re.compile(r'[a-z]+([A-Z]+[a-z0-9]*)*$')

_mypath = os.path.dirname(os.path.abspath(__file__))

class Version(object):
    def __init__(self, major=1, minor=0):
        self.major = major
        self.minor = minor

    def number(self):
        return (self.major << 8) + self.minor

    def __str__(self):
        return '%s.%s' % (self.major, self.minor)

    def __eq__(self, other):
        return (self.major, self.minor) == (other.major, other.minor)

    def __ne__(self, other):
        return (self.major, self.minor) != (other.major, other.minor)

    def __lt__(self, other):
        return (self.major, self.minor) <  (other.major, other.minor)

    def __le__(self, other):
        return (self.major, self.minor) <= (other.major, other.minor)

    def __gt__(self, other):
        return (self.major, self.minor) >  (other.major, other.minor)

    def __ge__(self, other):
        return (self.major, self.minor) <= (other.major, other.minor)

    def __hash__(self):
        return hash(self.major) * hash(self.minor)


class Type(streqobject):
    """
    Attributes:
        py_ctype   python ctypes type used as restype type and for structures
        py_argtype python type towards the C interface used as ctype argtype.
                   may be a converter containing a from_param class method
        py_ctor    callable to construct a python object from it's ctypes type value
                   mostly used for converting bytes (as a value of c_char_p) to str
                   can be "" in which case it acts as noop
    """
    def __init__(self, c_type, otf2_type, public_name='',
                 py_argtype=None, py_ctype=None, py_ctor="", **kwargs):
        self.c = c_type
        self.otf2 = otf2_type
        self.lower = otf2_type.lower()

        if py_ctype:
            self.py_ctype = py_ctype
        else:
            if self.c[:5] == 'OTF2_':
                self.py_ctype = self.c[5:]
            else:
                self.py_ctype = self.c

        if py_argtype:
            self.py_argtype = py_argtype
        else:
            self.py_argtype = py_ctype

        for key in kwargs:
            setattr(self, key, kwargs[key])

        # provide us as the c_type in the global symbols
        self.name = self.c
        if public_name is not None:
            if public_name:
                self.name = public_name
            globals()[self.name] = self
        self.lname = self.name[0].lower() + self.name[1:]

        types.append(self)

    def mutate(self, new_name):
        self.c_orig = self.c
        self.c = "OTF2_" + new_name
        self.py_underlying_ctype = self.py_ctype
        self.py_ctype = new_name
        self.py_argtype = new_name

    def size(self, attr=None):
        return 'sizeof( %s )' % (self.c,)

    def __str__(self):
        return str(self.c)


class _TypeAttributeFactory(object):
    def __init__(self, type, lname):
        self.type = type
        self.lname = lname

    def __call__(self, doc, name=None, **kwargs):
        if name is None:
            name = self.lname
        return _TypeAttribute(doc, self.type, name, **kwargs)

class RefType(Type):
    def __init__(self, *args, **kwargs):
        Type.__init__(self, *args, **kwargs)

class CompressedType(Type):
    def __init__(self, c_type, otf2_type, public_name='', attribute=None, **kwargs):
        Type.__init__(self, c_type, otf2_type, public_name, **kwargs)
        self.errorcheck_on_read = 'Invalid compression size.'
        if attribute:
            attr_name = attribute + 'Attribute'
            globals()[attr_name] = _TypeAttributeFactory(self,
                    attribute[0].lower() + attribute[1:])
            __all__.append(attr_name)

    def size(self, attr=None):
        return '%s + 1' % (Type.size(self, attr),)

class CompressedRefType(CompressedType, RefType):
    def __init__(self, *args, **kwargs):
        CompressedType.__init__(self, *args, **kwargs)

class StringType(Type):
    def __init__(self):
        Type.__init__(self, 'const char*', 'String', 'string_t', py_argtype='StrParam', py_ctype='ctypes.c_char_p')
        self.errorcheck_on_argument = 'Invalid string argument.'
        self.errorcheck_on_read = 'Can\'t read string from buffer.'

    def size(self, attr):
        return 'strlen( %s ) + 1' % (attr.name,)


class UnionType(Type):
    """
    ctypes does not support unions as (non-pointer) arguments.
    _arg_value / _arg_type is a crude and evil hack to guess
    the right argument type. This type must be correct in terms of
    size (maximum of the union) and
    calling convention (architecture specific, but for AMD64 int usually wins).
    Whenever used as a direct function argument (also applies to callbacks)
    this type should be used instead of the union. Otherwise you get undefined
    behavior.
    """
    def __init__(self, *args, **kwargs):
        Type.__init__(self, *args, **kwargs)


class AttributeValueType(UnionType):
    def __init__(self):
        UnionType.__init__(self, 'OTF2_AttributeValue', 'AttributeValue', 'OTF2_AttributeValue',
                           lower='attribute_value',
                           py_ctype='AttributeValue')
        self.errorcheck_on_read = 'Could not read attribute value.'
        globals()['AttributeValueAttribute'] = _TypeAttributeFactory(self, 'attributeValue')
        __all__.append('AttributeValueAttribute')

    # at most sizeof( uint64_t ) + 1
    def size(self, attr):
        return 'sizeof( %s ) + 1' % (attr.name,)

class IdMapType(Type):
    def __init__(self):
        Type.__init__(self, 'const OTF2_IdMap*', 'IdMap', 'OTF2_IdMap', lower='id_map', py_ctype='ctypes.POINTER(IdMap)')
        self.errorcheck_on_argument = 'Invalid IdMap argument.'
        self.errorcheck_on_read = 'Can\'t read IdMap.'

    def size(self, attr):
        return 'otf2_id_map_get_size( %s )' % (attr.name,)

_execfile(os.path.join(_mypath, 'otf2.types'))

class EnumAttributeFactory(object):
    def __init__(self, enum, enum_type=None):
        self.enum = enum
        self.type = enum.type
        if enum_type:
            self.type = enum_type

    def __call__(self, doc, name=None, **kwargs):
        if name is None:
            name = self.enum.lname
        if self.enum.attic \
                and ('attic' not in kwargs or kwargs['attic'] is None):
            raise AttributeError('Enum is attic since version %d.%d, but used as record attribute not in attic.' % (self.enum.attic.major, self.enum.attic.minor))
        return EnumAttribute(doc, self.enum, self.type, name, **kwargs)

class Alias(object):
    def __init__(self, doc, name, since=None, attic=None):
        self.name = name
        self.suffix = name
        if doc is None:
            doc = 'Default test string'
        self.doc = textwrap.dedent(doc)
        if len(self.doc) and self.doc[-1] == '\n':
            self.doc = self.doc[:-1]
        if since is not None:
            self.doc += '\n\n@since Version %s.' % (since,)
        self.deprecated_attribute = ''
        if attic is not None:
            self.doc += '\n\n@deprecated In version %s.' % (attic,)
            self.deprecated_attribute = ' __otf2_deprecated__'
    def __str__(self):
        return str(self.name)
    def c_decl(self, value, prefix, docprefix, width):
        return """
{0}/** @brief {1}
{0} */
{0}{2}{3} = {4}""".format(prefix, self.doxygendoc(docprefix, width),
                   self.name, self.deprecated_attribute, value)
    def doxygendoc(self, prefix = ' *  ', width = 70):
        return _doxygendoc(self.doc, prefix, width)

class EnumEntry(object):
    def __init__(self, doc, name, value=None, since=None, **kwargs):
        self.doc = textwrap.dedent(doc)
        if len(self.doc) and self.doc[-1] == '\n':
            self.doc = self.doc[:-1]
        self.name = name
        self.suffix = name
        self.value = value
        self.since = since
        self.aliases = []
        for alias in kwargs.pop('aliases', []):
            self.aliases.append(alias)

        for key in kwargs:
            setattr(self, key, kwargs[key])

        globals()[self.name] = self

    def prefixlen(self, length):
        self.suffix = self.name[length:]
        for alias in self.aliases:
            alias.suffix = str(alias)[length:]


    def doxygendoc(self, prefix=' *  ', width=70):
        return _doxygendoc(self.doc, prefix, width)

    def __str__(self):
        return str(self.name)

    def c_decl(self, prefix='    ', docprefix=' *  ', width=70):
        primary_decl = """{0}/** @brief {1}
{0} */
{0}{2} = {3}""".format(prefix, self.doxygendoc(prefix + docprefix, width), self.name, self.value)
        all_aliases = list(map(lambda alias: alias.c_decl(self.value, prefix, docprefix, width), self.aliases))
        all_aliases.insert(0, primary_decl)
        return ','.join(all_aliases)

class Enum(streqobject):
    def __init__(self, doc, name, *entries, **kwargs):
        self.since = kwargs.pop('since', Version())
        self.attic = kwargs.pop('attic', None)
        self.bitset_like = kwargs.pop('bitset_like', False)
        base_type = kwargs.pop('type', None)
        terminator = kwargs.pop('terminator', None)
        includes = kwargs.pop('includes', [])
        aliases = kwargs.pop('aliases', [])

        if _camel_case_name_re.match(name) is None:
            raise ValueError('Invalid enum name %s.' % (name,))

        # check for duplicate record name
        if name in enums:
            raise ValueError('A record/enum of name %s already exists.' % (name,))

        self.doc = textwrap.dedent(doc)
        if len(self.doc) and self.doc[-1] == '\n':
            self.doc = self.doc[:-1]
        self.name = name
        self.entries = []
        self.prefix = None

        if not base_type:
            base_type = uint8_t
        assert isinstance(base_type, Type), 'Invalid value type of \'type\' argument: %r' % (base_type,)
        self.type = copy.copy(base_type)
        self.type.mutate(self.name)

        assert isinstance(self.since, Version), 'Invalid value type for @since: %r' % (self.since,)
        self.latest_since = self.since

        for key in kwargs:
            setattr(self, key, kwargs[key])

        # convert camel-case to underscore
        res = re.subn(r'([A-Z]+)', r'_\1', self.name)
        # remove first _
        res = re.subn(r'^_', '', res[0])

        self.upper = res[0].upper()
        self.lower = res[0].lower()
        self.lname = self.name[0].lower() + self.name[1:]

        ids = set((-1,))
        for entry in entries:
            assert isinstance(entry, EnumEntry), 'Not of type EnumEntry'
            if entry.value is None:
                entry.value = max(ids) + 1
            assert entry.value not in ids, 'Duplicate enum value %s' % (entry.value,)
            if entry.since is None:
                entry.since = self.since
            if self.latest_since > entry.since:
                raise AttributeError('Enum entries must be added in monotonically increasing @since order: %s > %s' % (self.latest_since, entry.since))
            if self.latest_since < entry.since:
                self.latest_since = entry.since
            if self.since < entry.since:
                entry.doc += '\n\n@since Version %s.' % (entry.since,)
            ids.add(entry.value)
            self.entries.append(entry)

        self.prefix = os.path.commonprefix([x.name for x in self.entries])
        # prefix should be empty, or last character is an '_'
        if self.prefix != '' and self.prefix[-1] != '_':
            last = self.prefix.rfind('_')
            if last == -1:
                self.prefix = ''
            else:
                self.prefix = self.prefix[0:last+1]

        # includes get the common prefix
        for enum_name in includes:
            doc_extension = ''
            if isinstance(enum_name, tuple):
                enum_name, doc_extension = enum_name
            if doc_extension:
                doc_extension = '\n\n' + doc_extension
            assert enum_name in enums, 'Unknown enum in include statement: %s' % (enum_name,)
            enum = enums[enum_name]
            for entry in enum.entries:
                new_entry = EnumEntry(
                    entry.doc + doc_extension,
                    self.prefix + entry.suffix,
                    max(ids) + 1)
                ids.add(new_entry.value)
                self.entries.append(new_entry)

        # set the suffix name for all entries
        for entry in self.entries:
            entry.prefixlen(len(self.prefix))

        if terminator is not None and terminator is not True:
            assert isinstance(terminator, EnumEntry), 'Terminator is not of type EnumEntry'
        self.terminator = terminator

        if self.terminator is True:
            assert len(ids) > 1, 'Auto terminator without other entries.'
            self.terminator = EnumEntry(
                'Max entry.',
                self.prefix + 'MAX',
                max(ids) + 1)

        if self.terminator is not None:
            self.terminator.prefixlen(len(self.prefix))

        attr_name = name + 'Attribute'
        globals()[attr_name] = EnumAttributeFactory(self)
        __all__.append(attr_name)

        self.aliases = []
        for alias in aliases:
            alias_type = copy.copy(self.type)
            alias_type.mutate(alias)

            self.aliases.append(alias_type)

            attr_name = alias + 'Attribute'
            globals()[attr_name] = EnumAttributeFactory(self, alias_type)
            __all__.append(attr_name)

        if not self.attic:
            enums.append(self)
        else:
            attic_enums.append(self)

    def doxygendoc(self, prefix=' *  ', width=70):
        return _doxygendoc(self.doc, prefix, width)

    def __str__(self):
        return self.name

    def c_decl(self):
        entries = list(map(EnumEntry.c_decl, self.entries))
        if self.terminator:
            entries.append(self.terminator.c_decl())
        body = ',\n\n'.join(entries)
        return '''enum {0}_enum
{{
{1}
}};'''.format(self.type, body)

class GeneralEnum(Enum):
    def __init__(self, doc, name, *entries, **kwargs):
        Enum.__init__(self, doc, name, *entries, **kwargs)

class DefEnum(Enum):
    def __init__(self, doc, name, *entries, **kwargs):
        Enum.__init__(self, doc, name, *entries, **kwargs)

class EventEnum(Enum):
    def __init__(self, doc, name, *entries, **kwargs):
        Enum.__init__(self, doc, name, *entries, **kwargs)

_execfile(os.path.join(_mypath, 'otf2.enums'))

class Attribute(object):
    """Represents an attribute for any record type."""

    def __init__(self, doc, type, name, **kwargs):
        self.attic = kwargs.pop('attic', None);
        self.since = kwargs.pop('since', None);
        self.doc = textwrap.dedent(doc)
        if len(self.doc) and self.doc[-1] == '\n':
            self.doc = self.doc[:-1]
        self.type = type
        self.set_name(name)
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def size(self):
        return self.type.size(self)

    def set_name(self, name):
        if _lower_camel_case_name_re.match(name) is None:
            raise ValueError('Invalid attribute name %s.' % (name,))

        self.name = name
        res = re.subn(r'([A-Z]+)', r'_\1', self.name)
        self.upper = res[0].upper()
        self.lower = res[0].lower()

    def doxygendoc(self, prefix=' *  ', width=70):
        return _doxygendoc(self.doc, prefix, width)

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.name)

class ArrayAttribute(Attribute):
    """Represents an array attribute which has multiple child attributes for
    which this attribute is the count attribute."""

    def __init__(self, doc, type, name, *arrayAttributes, **kwargs):
        Attribute.__init__(self, doc, type, name, **kwargs)
        assert len(arrayAttributes) >= 1, 'Need at least one value attribute member'
        self.array_attributes = []
        for array_attr in arrayAttributes:
            assert not isinstance(array_attr, ArrayAttribute), 'Recursive array attributes are not support.'
            self.array_attributes.append(array_attr)

    def __repr__(self):
        return '<%s: %s[%s]>' % (
            self.__class__.__name__,
            self.name,
            ','.join([repr(x) for x in self.array_attributes]))

class _TypeAttribute(Attribute):
    def __init__(self, doc, type, name, **kwargs):
        Attribute.__init__(self, doc, type, name, **kwargs)

class DefRefAttribute(Attribute):
    """Represents an attribute for event or definition records, that reference
     a definition."""

    def __init__(self, doc, type, name, def_ref, **kwargs):
        self.mapping_type = kwargs.pop('mapping_type', None)
        Attribute.__init__(self, doc, type, name, **kwargs)
        self.def_ref = def_ref

    def __repr__(self):
        return '<%sAttribute: %s>' % (self.def_ref.name, self.name)

class SelfRefAttribute(object):
    """Represents an attribute for definition records, that references itself."""

    def __init__(self, doc, name=None, since=None):
        self.doc = doc
        self.name = name
        self.since = since

class TimeStampAttribute(Attribute):
    """\
    Represents an timestamp attribute for event records only.
    The clock offsets will be applied to timestamp attributes.
    """
    def __init__(self, doc, name, **kwargs):
        Attribute.__init__(self, doc, OTF2_TimeStamp, name, **kwargs)
        self.apply_clock_offset = True

class DefRefAttributeFactory(object):
    def __init__(self, type, def_ref, mapping_type=None):
        self.type = type
        self.def_ref = def_ref
        self.mapping_type = mapping_type

    def __call__(self, doc, name=None, **kwargs):
        if name is None:
            name = self.def_ref.lname
        return DefRefAttribute(doc, self.type, name,
                self.def_ref,
                mapping_type=self.mapping_type,
                **kwargs)

class EnumAttribute(Attribute):
    def __init__(self, doc, enum, enum_type, name, **kwargs):
        self.enum = enum
        Attribute.__init__(self, doc, enum_type, name, **kwargs)

class Record(streqobject):
    def __init__(self, doc, id, name, since=Version(), global_name=None, **kwargs):
        if _camel_case_name_re.match(name) is None:
            raise ValueError('Invalid record name %s.' % (name,))

        self.id = id
        self.name = name
        assert isinstance(since, Version), 'Invalid value type for @since: %r' % (since,)
        self.since = since
        self.deprecated = kwargs.pop('deprecated', None)
        assert self.deprecated is None or isinstance(self.deprecated, Version), 'Invalid value type for @deprecated: %r' % (self.deprecated,)
        self.versions = [self.since]
        self.max_attr_len = 0

        for key in kwargs:
            setattr(self, key, kwargs[key])

        # convert camel-case to underscore
        res = re.subn(r'([A-Z]+)', r'_\1', name)
        # remove first _
        res = re.subn(r'^_', '', res[0])

        self.upper = res[0].upper()
        self.lower = res[0].lower()
        self.lname = name[0].lower() + name[1:]
        self.doc = textwrap.dedent(doc)
        if len(self.doc) and self.doc[-1] == '\n':
            self.doc = self.doc[:-1]
        self.attribute_by_name = {}
        self.attributes = []
        self.array_attributes = []
        self.all_attributes = []
        self.all_array_attributes = []
        self.all_value_attributes = []
        self.record_attributes = []
        self.record_attributes_by_version = {}
        self.attic_attributes_by_version = {}

        if global_name is None:
            global_name = self.name
        globals()[global_name + 'Record'] = self

    def _addAttribute(self, attr):
        assert isinstance(attr, Attribute), 'Only accept attributes of type Attribute: %s' % type(attr)
        if attr.since is not None and not isinstance(attr.since, Version):
            raise AttributeError('Invalid value type for @since: %r' % (attr.since,))
        if attr.attic is not None and not isinstance(attr.attic, Version):
            raise AttributeError('Invalid value type for @attic: %r' % (attr.attic,))
        if attr.since is None:
            attr.since = self.since
        if self.since < attr.since:
            attr.doc += '\nSince version %s.' % (attr.since,)
            if attr.since not in self.versions:
                self.versions.append(attr.since)
                self.versions.sort()
        if attr.name in self.attribute_by_name:
            raise ValueError('An attribute of this name already exist: %s (%s).' % (attr.name, ','.join(self.attribute_by_name)))
        self.attribute_by_name[attr.name] = attr
        self.record_attributes.append(attr)
        self.record_attributes.sort(key=lambda x: x.since)
        if attr.since not in self.record_attributes_by_version:
            self.record_attributes_by_version[attr.since] = []
        self.record_attributes_by_version[attr.since].append(attr)
        if not attr.attic:
            self.attributes.append(attr)
            self.all_attributes.append(attr)
        else:
            if attr.attic not in self.attic_attributes_by_version:
                self.attic_attributes_by_version[attr.attic] = []
            self.attic_attributes_by_version[attr.attic].append(attr)
        if len(attr.name) > self.max_attr_len:
            self.max_attr_len = len(attr.name)
        if isinstance(attr, ArrayAttribute):
            self.array_attributes.append(attr)
            for array_attr in attr.array_attributes:
                if array_attr.since is not None:
                    raise AttributeError('Array attribute members should not have a @since added')
                array_attr.since = attr.since
                if self.since < attr.since:
                    attr.doc += '\nSince version %s.' % (attr.since,)
                array_attr.set_name(pluralize(array_attr.name))
                if array_attr.name in self.attribute_by_name:
                    raise ValueError('An attribute of this name already exist: %s.' % (array_attr.name,))
                self.attribute_by_name[array_attr.name] = array_attr
                self.all_attributes.append(array_attr)
                self.all_array_attributes.append(array_attr)
                self.all_value_attributes.append(array_attr)
                if len(array_attr.name) > self.max_attr_len:
                    self.max_attr_len = len(array_attr.name)
        else:
            self.all_value_attributes.append(attr)
    def doxygendoc(self, prefix=' *  ', width=70):
        return _doxygendoc(self.doc, prefix, width)


    def formatargs(self, format_default, format_refs=None, leading_comma=True):
        args = []
        for attr in self.attributes:
            format_args = {
                'attr': attr,
                'lower': attr.lower,
                'name': attr.name,
                'upper': attr.upper,
            }
            if format_refs and isinstance(attr, DefRefAttribute):
                arg = format_refs.format(def_ref=attr.def_ref, **format_args)
            else:
                arg = format_default.format(**format_args)
            args.append(arg)

        sep = ', '
        if leading_comma:
            return sep + sep.join(args)
        else:
            return sep.join(args)


    def py_constructargs(self, leading_comma=True):
        args = []
        for attr in self.attributes:
            if isinstance(attr, DefRefAttribute):
                arg = "registry.{}s[{}]".format(attr.def_ref.lower, attr.lower)
            else:
                arg = attr.lower
            args.append(arg)

        sep = ', '
        if leading_comma:
            return sep + sep.join(args)
        else:
            return sep.join(args)

    def py_callargs(self, *args, **kwargs):
        return self.py_funcargs(*args, callargs=True, **kwargs)

    def py_funcargs(self,
                    attributes=None,
                    case='name',
                    prefix='',
                    leading_comma=True,
                    resolve_array_attributes=True,
                    pass_by_ref=None,
                    callargs=False):
        """\
        Generate a argument list for python functions.
        Can be used for calling and for signature

        Output just the argument names
        """

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attrs:
            arg_name = getattr(attr, case)
            if callargs and isinstance(attr.type, UnionType):
                arg_name = "%s._arg_value" % arg_name
            args.append('%s' % arg_name)
            if resolve_array_attributes and isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    if pass_by_ref:
                        args.append('ctypes.byref(%s)' % (getattr(array_attr, case)))
                    else:
                        args.append('%s' % (getattr(array_attr, case)))
        sep = ', %s' % (prefix,)
        if leading_comma:
            return sep + sep.join(args)
        else:
            return sep.join(args)

    def py_funcargtypes(self,
                        attributes=None,
                        case='name',
                        prefix='',
                        leading_comma=True,
                        resolve_array_attributes=True,
                        argtypes=True):
        """\
        Generate a function argument type list for python ctypes.
        argtypes=True:  Use in argtypes = ...
        argtypes=False: Use for callback function pointer types
        """

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attrs:
            if argtypes:
                py_type = attr.type.py_argtype
            else:
                py_type = attr.type.py_ctype

            if isinstance(attr.type, UnionType):
                args.append('%s._arg_type' % py_type)
            else:
                args.append('%s' % py_type)

            if resolve_array_attributes and isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    args.append('ctypes.POINTER(%s)' % (str(array_attr.type.py_ctype)))
        sep = ', %s' % (prefix,)
        if leading_comma:
            return sep + sep.join(args)
        else:
            sep.join(args)

    def funcargs(self,
                 attributes=None,
                 case='name',
                 prefix='',
                 leading_comma=True,
                 resolve_array_attributes=True):
        """\
        Generate an function prototype argument list.

        Empty for records without any attributes. Otherwise the attributes are
        separated by `,\n%(prefix)`, including a first comma if `leading_comma`
        is true.
        """

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attrs:
            args.append('%s %s' % (attr.type, getattr(attr, case)))
            if resolve_array_attributes and isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    args.append('const %s* %s' % (array_attr.type, getattr(array_attr, case)))
        sep = ',\n%s' % (prefix,)
        if leading_comma:
            return sep + sep.join(args)
        else:
            return sep.join(args)

    def callargs(self,
                 attributes=None,
                 case='name',
                 argprefix='',
                 prefix='',
                 leading_comma=True,
                 resolve_array_attributes=True):
        """\
        Generates a function call argument list.

        Empty for records with no attributes. Otherwise the attributes are
        separated by `,\n%(prefix)`, including a first comma if `leading_comma`
        is true.
        """

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attrs:
            args.append('%s%s' % (argprefix, getattr(attr, case)))
            if resolve_array_attributes and isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    args.append('%s%s' % (argprefix, getattr(array_attr, case)))
        sep = ',\n%s' % (prefix,)
        if leading_comma:
            return sep + sep.join(args)
        else:
            return sep.join(args)

    def printargs(self,
                  case='name',
                  argprefix='',
                  prefix=''):
        """\
        print argument list without linebreaks.

        Empty for records with no attributes. Otherwise the attributes are
        separated by `%(prefix)`.
        """

        if len(self.attributes) == 0:
            return ''

        args = []
        for attr in self.all_attributes:
            args.append('%s%s' % (argprefix, getattr(attr, case)))
        sep = '%s ' % (prefix,)
        return sep + sep.join(args)


    def structargs(self,
                   attributes=None,
                   case='name',
                   prefix='    '):
        """Generates the list of members of a struct for the attributes.
        (include Doxygen-style documentation)."""

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attributes:
            args.append('%s %s;' % (attr.type, getattr(attr, case)))
            if isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    args.append('%s* %s;' % (array_attr.type, getattr(array_attr, case)))
        return ('\n%s' % (prefix,)).join(args)

    def pystructargs(self,
                     attributes=None,
                     case='name',
                     prefix='   '):
        """Generates the list of members of a struct for the attributes.
        As Python ctypes classes. """

        attrs = self.attributes
        if attributes is not None:
            attrs = attributes

        if len(attrs) == 0:
            return ''

        args = []
        for attr in attributes:
            args.append('("%s", %s),' % (getattr(attr, case), str(attr.type.py_ctype)))
            if isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    args.append('("%s", ctypes.POINTER(%s)),' %
                            (getattr(array_attr, case), str(array_attr.type.py_ctype)))
        return ('\n%s' % (prefix,)).join(args)

    def doxygenargs(self,
                    attributes=None,
                    case='name',
                    prefix=' *  ',
                    initial_prefix='\n *  ',
                    resolve_array_attributes=True,
                    prepend=list()):
        """Generates doxygen-style @param list for the attributes."""

        attrs = self.attributes
        max_attr_len = self.max_attr_len
        if attributes is not None:
            max_attr_len = 0
            attrs = attributes
            for attr in attrs:
                attr_name = getattr(attr, case)
                if len(attr_name) > max_attr_len:
                    max_attr_len = len(attr_name)

        for parameter, doc in prepend:
            if len(parameter) > max_attr_len:
                max_attr_len = len(parameter)

        if len(attrs) == 0 and len(prepend) == 0:
            return ''

        result = []
        for parameter, doc in prepend:
            result.append(
                textwrap.fill(
                    doc,
                    width=77,
                    initial_indent='@param %-*s ' % (max_attr_len, parameter),
                    subsequent_indent='%s%s' % (prefix, ' ' * (8 + max_attr_len)),
                    break_long_words=False))

        for attr in attrs:
            result.append(
                textwrap.fill(
                    attr.doc,
                    width=77,
                    initial_indent='@param %-*s ' % (max_attr_len, getattr(attr, case)),
                    subsequent_indent='%s%s' % (prefix, ' ' * (8 + max_attr_len)),
                    break_long_words=False))
            if resolve_array_attributes and isinstance(attr, ArrayAttribute):
                for array_attr in attr.array_attributes:
                    result.append(
                        textwrap.fill(
                            array_attr.doc,
                            width=77,
                            initial_indent='@param %-*s ' % (max_attr_len, getattr(array_attr, case)),
                            subsequent_indent='%s%s' % (prefix, ' ' * (8 + max_attr_len)),
                            break_long_words=False))
        return initial_prefix + ('\n%s' % (prefix,)).join(result)

    def deprecate_decl(self):
        if self.deprecated is not None:
            return ' __otf2_deprecated__'
        return ''

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return '<%s: %s(%s)>' % (
            self.__class__.__name__,
            self.name,
            ','.join([repr(x) for x in self.attributes]))


class DefRecord(Record):
    """\
    Declare a definition record.

    Attributes:
        self_attr: The attribute for the unique ID of this definition.
                   Maybe None if this definition does not have a unique ID.

        siblings: List of definitions, which share the same ID space.
                  If len(siblings) == 1 this definition is a singleton.

        supplements: List of definitions, which are a supplement to this one.
    """
    def __init__(self, doc, id, name, self_type, *attributes, **kwargs):
        mapping_type = kwargs.pop('mapping_type', None)
        self.archive_callback = kwargs.pop('archive_callback', None)
        self.ref_name = kwargs.pop('ref_name', True)

        # all children defs, which are supplementary defs to this one
        self.supplements = []

        # check for duplicate record name
        if name in defs:
            raise ValueError('A definition named %s already exists.' % (name,))

        Record.__init__(self, doc, id, name, **kwargs)

        # if we have a type, than add an implicit identifier attribute
        self.self_attr = None
        # list of defs which share the same ID space
        # if self.siblings[0] == self holds, self is the first born
        self.siblings = [self]
        if self_type:
            if isinstance(self_type, Record):
                assert isinstance(self_type, DefRecord), 'Only definition records allowed.'
                first_born = self_type
                # the referenced def sibling needs to be the first born
                assert first_born.siblings[0] == first_born, \
                    'The referenced definition sibling needs to be the first born.'
                first_born.siblings.append(self)
                self.siblings = first_born.siblings
                self.ref_name = first_born.ref_name
                self.ref_upper = first_born.ref_upper
                self.ref_lower = first_born.ref_lower
                self.ref_lname = first_born.ref_lname
                self.self_attr = copy.copy(first_born.self_attr)
                self.self_attr.since = self.since
                self_type = first_born.self_attr.type
            elif isinstance(self_type, RefType):
                self_type = copy.copy(self_type)

                # determine our name to reference this definition
                if self.ref_name is True:
                    self.ref_name  = self.name
                if self.ref_name is not None:
                    if _camel_case_name_re.match(self.ref_name) is None:
                        raise ValueError('Invalid record def ref name %s.' % (self.ref_name,))
                    self.ref_lname = self.ref_name[0].lower() + self.ref_name[1:]

                    # convert camel-case to underscore
                    res = re.subn(r'([A-Z]+)', r'_\1', self.ref_name)
                    # remove first _
                    res = re.subn(r'^_', '', res[0])

                    self.ref_upper = res[0].upper()
                    self.ref_lower = res[0].lower()

                    self_type.mutate(self.ref_name + 'Ref')

                    if mapping_type:
                        setattr(mapping_type, 'def', self)

                    self.self_attr = DefRefAttribute(
                        'The unique identifier for this @eref{%s} definition.' % (self.name,),
                        self_type,
                        'self',
                        self,
                        mapping_type=mapping_type)

                    # provide a <name>Attribute class, so that other defs and events can
                    # reference us
                    attr_name = self.ref_name + 'Attribute'
                    globals()[attr_name] = DefRefAttributeFactory(self_type, self, mapping_type)
                    __all__.append(attr_name)
            else:
                raise TypeError('Unexpected type of parameter: %s' % (str(self_type),))

            # prepend to all other attributes
            self._addAttribute(self.self_attr)

        else:
            assert not mapping_type, 'Definition record with mapping_type, but no identifier attribute.'

        def add_def_ref_doc(attr):
            if isinstance(attr, DefRefAttribute):
                defs = '}, or a @eref{'.join([x.name for x in attr.def_ref.siblings])
                attr.doc += '\nReferences a @eref{%s} definition.' % (defs,)

        # add all attributes, but catch SelfRefAttribute and resolve them
        # to our self
        for attr in attributes:
            assert not isinstance(attr, TimeStampAttribute), 'TimeStamp attribute used in definition record.'
            if isinstance(attr, SelfRefAttribute):
                assert self.self_attr is not None, 'Self referencing attribute without id.'
                if attr.name is None:
                    attr.name = 'parent'
                attr = DefRefAttribute(attr.doc, self_type, attr.name, self, since=attr.since)
            elif isinstance(attr, ArrayAttribute):
                new_array_attributes = []
                for array_attr in attr.array_attributes:
                    assert not isinstance(array_attr, TimeStampAttribute), 'TimeStamp attribute used in definition record.'
                    if isinstance(array_attr, SelfRefAttribute):
                        assert self.self_attr is not None, 'Self referencing attribute without id.'
                        if array_attr.name is None:
                            attr.name = 'parent'
                        array_attr = DefRefAttribute(array_attr.doc, self_type, array_attr.name, self, since=attr.since)
                    add_def_ref_doc(array_attr)
                    new_array_attributes.append(array_attr)
                attr.array_attributes = new_array_attributes
            add_def_ref_doc(attr)
            self._addAttribute(attr)

        defs.append(self)
        defs.sort(key=lambda x: x.id)

class LocalOnlyDefRecord(DefRecord):
    def __init__(self, doc, id, name, *attributes, **kwargs):
        self_type = kwargs.pop('type', None)
        self.archive_location_callback = kwargs.pop('archive_location_callback', None)
        DefRecord.__init__(self, doc, id, name, self_type, *attributes, **kwargs)

class GlobalOnlyDefRecord(DefRecord):
    def __init__(self, doc, id, name, *attributes, **kwargs):
        self_type = kwargs.pop('type', None)
        DefRecord.__init__(self, doc, id, name, self_type, *attributes, **kwargs)

class SuppDefRecord(DefRecord):
    def __init__(self, doc, id, name, parent_def, *attributes, **kwargs):
        assert isinstance(parent_def, DefRecord), 'Parent must be a definition: %r' % (parent_def,)
        self.parent_def = parent_def
        assert self.parent_def.self_attr, 'Parent must be a referable definition: %r' % (self.parent_def,)
        new_attributes = [
                DefRefAttribute(
                        'Parent @eref{%s} definition to which this one is a supplementary definition.' % ('}, or @eref{'.join([x.name for x in self.parent_def.siblings]),),
                        self.parent_def.self_attr.type,
                        self.parent_def.ref_lname,
                        self.parent_def)] + list(attributes)
        DefRecord.__init__(self, doc, id, name, None, *new_attributes, **kwargs)
        self.parent_def.supplements.append(self)

_execfile(os.path.join(_mypath, 'otf2.defs'))

class EventRecord(Record):
    def __init__(self, doc, id, name, *attributes, **kwargs):
        # check for duplicate record name
        if name in events:
            raise ValueError('A event named %s already exists.' % (name,))

        Record.__init__(self, doc, id, name, **kwargs)

        # do not permit event records after 1.1 to be singleton
        # they can't be handled when skipping records and there payload
        # is an UNDEFINED value (ie. 0xff) which conflicts
        # with the 0xff record length byte indicating a record
        # length >= 256
        self.singleton = len(attributes) == 1 and self.since <= Version(1, 1)
        for attr in attributes:
            if isinstance(attr, DefRefAttribute):
                if not attr.mapping_type and not isinstance(attr.def_ref, GlobalOnlyDefRecord):
                    raise ValueError('Definition reference attribute does not have a mapping type. Needed when used in event records.')
                defs = '}, or a @eref{'.join([x.name for x in attr.def_ref.siblings])
                if attr.mapping_type:
                    attr.doc += '\nReferences a @eref{%s} definition and will be mapped to the global definition if a mapping table of type @eref{%s} is available.' % (defs, attr.mapping_type)
                else:
                    attr.doc += '\nReferences a @eref{%s} definition.' % (defs,)
            if isinstance(attr, ArrayAttribute):
                # having an array attribute destroys the singleton property
                self.singleton = False
            if not isinstance(attr.type, CompressedType):
                # having an attribute which will not be compressed destroys the
                # singleton property
                self.singleton = False
            self._addAttribute(attr)

        events.append(self)
        events.sort(key=lambda x: x.id)

_execfile(os.path.join(_mypath, 'otf2.events'))


class SnapshotRecord(Record):
    def __init__(self, doc, id, name_or_origin, *attributes, **kwargs):
        since = kwargs.pop('since', Version(1,2))

        if isinstance(name_or_origin, EventRecord):
            assert len(attributes) == 0, 'No attributes allowed when cloning an event record.'
            derived_event = True
            name = name_or_origin.name
            self.snap_name = name_or_origin.name + 'Snap'
            attributes = copy.deepcopy(name_or_origin.attributes)
            setattr(name_or_origin, 'snapshot_record', self)
            self.origin = name_or_origin
            for attr in attributes:
                if attr.since < since:
                    attr.since = since
                if isinstance(attr, ArrayAttribute):
                    for array_attr in attr.array_attributes:
                        array_attr.since = None
        else:
            assert isinstance(name_or_origin, str), 'Expecting the name of the record.'
            derived_event = False
            name = name_or_origin
            self.snap_name = name_or_origin
            attributes = list(attributes)

        # check for duplicate record name
        if name in snaps:
            raise ValueError('A snapshot event named %s already exists.' % (name,))

        Record.__init__(self, doc, id, name,
                since=since, global_name=self.snap_name, **kwargs)

        if derived_event:
            attributes.insert(0, TimeStampAttribute(
                    """The original time this event happened.""",
                    'origEventTime'))

        for attr in attributes:
            self._addAttribute(attr)

        snaps.append(self)
        snaps.sort(key=lambda x: x.id)

_execfile(os.path.join(_mypath, 'otf2.snaps'))




''' Generated by OTF2 Template Engine '''

import ctypes
import traceback
import sys

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .AttributeList import AttributeList
from .Definitions import *
from .Events import *
from .Callbacks import callback_wrapper


class GlobalSnapReaderCallbacks(ctypes.Structure):
    pass

def GlobalSnapReaderCallbacks_New():
    c_New = conf.lib.OTF2_GlobalSnapReaderCallbacks_New
    c_New.argtypes = []
    c_New.restype = ctypes.POINTER(GlobalSnapReaderCallbacks)
    # NOTE: Do not errcheck here. This function returns a value, not an error code
    return c_New()

def GlobalSnapReaderCallbacks_Delete(globalSnapReaderCallbacks):
    c_Delete = conf.lib.OTF2_GlobalSnapReaderCallbacks_Delete
    c_Delete.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks) ]
    c_Delete.restype = None
    return c_Delete(globalSnapReaderCallbacks)

def GlobalSnapReaderCallbacks_Clear(globalSnapReaderCallbacks):
    c_Clear = conf.lib.OTF2_GlobalSnapReaderCallbacks_Clear
    c_Clear.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks) ]
    c_Clear.restype = None
    return c_Clear(globalSnapReaderCallbacks)

def _callback_wrapper(type, func, convert_args=None):
    def wrapper(location, time, userData, attribute_list, *args):
        if userData:
            py_userData = ctypes.cast(userData, ctypes.py_object).value
        else:
            py_userData = None
        try:
            if convert_args is not None:
                args = convert_args(*args)
            if not attribute_list:
                attribute_list = None
            ret = func(location, time, py_userData, attribute_list, *args)
            if ret is None:
                ret = CALLBACK_SUCCESS
        except:
            sys.stderr.write("An unhandled python exception has occurred in an "
                             "OTF2_GlobalSnapReaderCallback:\n")
            sys.stderr.write(traceback.format_exc())
            ret = CALLBACK_ERROR
        return ret.value
    return callback_wrapper(func, wrapper, type)

_GlobalSnapReaderCallback_FP_Unknown = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList))

_GlobalSnapReaderCallback_FP_SnapshotStart = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_SnapshotEnd = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MeasurementOnOff = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, MeasurementMode)

_GlobalSnapReaderCallback_FP_Enter = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, RegionRef)

_GlobalSnapReaderCallback_FP_MpiSend = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiIsend = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiIsendComplete = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiRecv = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiIrecvRequest = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiIrecv = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_MpiCollectiveBegin = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp)

_GlobalSnapReaderCallback_FP_MpiCollectiveEnd = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, CollectiveOp, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_OmpFork = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32)

_GlobalSnapReaderCallback_FP_OmpAcquireLock = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint32, ctypes.c_uint32)

_GlobalSnapReaderCallback_FP_OmpTaskCreate = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_OmpTaskSwitch = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64)

_GlobalSnapReaderCallback_FP_Metric = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, MetricRef, ctypes.c_uint8, ctypes.POINTER(Type), ctypes.POINTER(MetricValue))

_GlobalSnapReaderCallback_FP_ParameterString = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ParameterRef, StringRef)

_GlobalSnapReaderCallback_FP_ParameterInt = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ParameterRef, ctypes.c_int64)

_GlobalSnapReaderCallback_FP_ParameterUnsignedInt = ctypes.CFUNCTYPE(CallbackCode,
        LocationRef, TimeStamp, ctypes.c_void_p,
        ctypes.POINTER(AttributeList), TimeStamp, ParameterRef, ctypes.c_uint64)

def GlobalSnapReaderCallbacks_SetUnknownCallback(globalSnapReaderCallbacks, unknownCallback):
    c_SetUnknownCallback = conf.lib.OTF2_GlobalSnapReaderCallback_SetUnknownCallback
    c_SetUnknownCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_Unknown ]
    c_SetUnknownCallback.restype = ErrorCode
    c_SetUnknownCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_Unknown,
                                         unknownCallback)
    c_SetUnknownCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetSnapshotStartCallback(globalSnapReaderCallbacks, snapshotStartCallback):
    c_SetSnapshotStartCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetSnapshotStartCallback
    c_SetSnapshotStartCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_SnapshotStart ]
    c_SetSnapshotStartCallback.restype = ErrorCode
    c_SetSnapshotStartCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_snapshotStartCallback,
                                         snapshotStartCallback)
    c_SetSnapshotStartCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetSnapshotEndCallback(globalSnapReaderCallbacks, snapshotEndCallback):
    c_SetSnapshotEndCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetSnapshotEndCallback
    c_SetSnapshotEndCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_SnapshotEnd ]
    c_SetSnapshotEndCallback.restype = ErrorCode
    c_SetSnapshotEndCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_snapshotEndCallback,
                                         snapshotEndCallback)
    c_SetSnapshotEndCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMeasurementOnOffCallback(globalSnapReaderCallbacks, measurementOnOffCallback):
    c_SetMeasurementOnOffCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMeasurementOnOffCallback
    c_SetMeasurementOnOffCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MeasurementOnOff ]
    c_SetMeasurementOnOffCallback.restype = ErrorCode
    c_SetMeasurementOnOffCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_measurementOnOffCallback,
                                         measurementOnOffCallback)
    c_SetMeasurementOnOffCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetEnterCallback(globalSnapReaderCallbacks, enterCallback):
    c_SetEnterCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetEnterCallback
    c_SetEnterCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_Enter ]
    c_SetEnterCallback.restype = ErrorCode
    c_SetEnterCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_enterCallback,
                                         enterCallback)
    c_SetEnterCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiSendCallback(globalSnapReaderCallbacks, mpiSendCallback):
    c_SetMpiSendCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiSendCallback
    c_SetMpiSendCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiSend ]
    c_SetMpiSendCallback.restype = ErrorCode
    c_SetMpiSendCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiSendCallback,
                                         mpiSendCallback)
    c_SetMpiSendCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiIsendCallback(globalSnapReaderCallbacks, mpiIsendCallback):
    c_SetMpiIsendCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiIsendCallback
    c_SetMpiIsendCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiIsend ]
    c_SetMpiIsendCallback.restype = ErrorCode
    c_SetMpiIsendCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiIsendCallback,
                                         mpiIsendCallback)
    c_SetMpiIsendCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiIsendCompleteCallback(globalSnapReaderCallbacks, mpiIsendCompleteCallback):
    c_SetMpiIsendCompleteCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiIsendCompleteCallback
    c_SetMpiIsendCompleteCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiIsendComplete ]
    c_SetMpiIsendCompleteCallback.restype = ErrorCode
    c_SetMpiIsendCompleteCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiIsendCompleteCallback,
                                         mpiIsendCompleteCallback)
    c_SetMpiIsendCompleteCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiRecvCallback(globalSnapReaderCallbacks, mpiRecvCallback):
    c_SetMpiRecvCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiRecvCallback
    c_SetMpiRecvCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiRecv ]
    c_SetMpiRecvCallback.restype = ErrorCode
    c_SetMpiRecvCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiRecvCallback,
                                         mpiRecvCallback)
    c_SetMpiRecvCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiIrecvRequestCallback(globalSnapReaderCallbacks, mpiIrecvRequestCallback):
    c_SetMpiIrecvRequestCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiIrecvRequestCallback
    c_SetMpiIrecvRequestCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiIrecvRequest ]
    c_SetMpiIrecvRequestCallback.restype = ErrorCode
    c_SetMpiIrecvRequestCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiIrecvRequestCallback,
                                         mpiIrecvRequestCallback)
    c_SetMpiIrecvRequestCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiIrecvCallback(globalSnapReaderCallbacks, mpiIrecvCallback):
    c_SetMpiIrecvCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiIrecvCallback
    c_SetMpiIrecvCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiIrecv ]
    c_SetMpiIrecvCallback.restype = ErrorCode
    c_SetMpiIrecvCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiIrecvCallback,
                                         mpiIrecvCallback)
    c_SetMpiIrecvCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiCollectiveBeginCallback(globalSnapReaderCallbacks, mpiCollectiveBeginCallback):
    c_SetMpiCollectiveBeginCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiCollectiveBeginCallback
    c_SetMpiCollectiveBeginCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiCollectiveBegin ]
    c_SetMpiCollectiveBeginCallback.restype = ErrorCode
    c_SetMpiCollectiveBeginCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiCollectiveBeginCallback,
                                         mpiCollectiveBeginCallback)
    c_SetMpiCollectiveBeginCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMpiCollectiveEndCallback(globalSnapReaderCallbacks, mpiCollectiveEndCallback):
    c_SetMpiCollectiveEndCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMpiCollectiveEndCallback
    c_SetMpiCollectiveEndCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_MpiCollectiveEnd ]
    c_SetMpiCollectiveEndCallback.restype = ErrorCode
    c_SetMpiCollectiveEndCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_mpiCollectiveEndCallback,
                                         mpiCollectiveEndCallback)
    c_SetMpiCollectiveEndCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetOmpForkCallback(globalSnapReaderCallbacks, ompForkCallback):
    c_SetOmpForkCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetOmpForkCallback
    c_SetOmpForkCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_OmpFork ]
    c_SetOmpForkCallback.restype = ErrorCode
    c_SetOmpForkCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_ompForkCallback,
                                         ompForkCallback)
    c_SetOmpForkCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetOmpAcquireLockCallback(globalSnapReaderCallbacks, ompAcquireLockCallback):
    c_SetOmpAcquireLockCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetOmpAcquireLockCallback
    c_SetOmpAcquireLockCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_OmpAcquireLock ]
    c_SetOmpAcquireLockCallback.restype = ErrorCode
    c_SetOmpAcquireLockCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_ompAcquireLockCallback,
                                         ompAcquireLockCallback)
    c_SetOmpAcquireLockCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetOmpTaskCreateCallback(globalSnapReaderCallbacks, ompTaskCreateCallback):
    c_SetOmpTaskCreateCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetOmpTaskCreateCallback
    c_SetOmpTaskCreateCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_OmpTaskCreate ]
    c_SetOmpTaskCreateCallback.restype = ErrorCode
    c_SetOmpTaskCreateCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_ompTaskCreateCallback,
                                         ompTaskCreateCallback)
    c_SetOmpTaskCreateCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetOmpTaskSwitchCallback(globalSnapReaderCallbacks, ompTaskSwitchCallback):
    c_SetOmpTaskSwitchCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetOmpTaskSwitchCallback
    c_SetOmpTaskSwitchCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_OmpTaskSwitch ]
    c_SetOmpTaskSwitchCallback.restype = ErrorCode
    c_SetOmpTaskSwitchCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_ompTaskSwitchCallback,
                                         ompTaskSwitchCallback)
    c_SetOmpTaskSwitchCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetMetricCallback(globalSnapReaderCallbacks, metricCallback):
    c_SetMetricCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetMetricCallback
    c_SetMetricCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_Metric ]
    c_SetMetricCallback.restype = ErrorCode
    c_SetMetricCallback.errcheck = HandleErrorCode
    def convert_args_metric(metric, number_of_metrics, type_ids_array, metric_values_array):
        type_ids = [type_ids_array[i] for i in range(number_of_metrics)]
        metric_values = [metric_values_array[i] for i in range(number_of_metrics)]
        return metric, type_ids, metric_values
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_Metric,
                                         metricCallback, convert_args_metric)
    c_SetMetricCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetParameterStringCallback(globalSnapReaderCallbacks, parameterStringCallback):
    c_SetParameterStringCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetParameterStringCallback
    c_SetParameterStringCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_ParameterString ]
    c_SetParameterStringCallback.restype = ErrorCode
    c_SetParameterStringCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_parameterStringCallback,
                                         parameterStringCallback)
    c_SetParameterStringCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetParameterIntCallback(globalSnapReaderCallbacks, parameterIntCallback):
    c_SetParameterIntCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetParameterIntCallback
    c_SetParameterIntCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_ParameterInt ]
    c_SetParameterIntCallback.restype = ErrorCode
    c_SetParameterIntCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_parameterIntCallback,
                                         parameterIntCallback)
    c_SetParameterIntCallback(globalSnapReaderCallbacks, wrapped_callback)

def GlobalSnapReaderCallbacks_SetParameterUnsignedIntCallback(globalSnapReaderCallbacks, parameterUnsignedIntCallback):
    c_SetParameterUnsignedIntCallback = conf.lib.OTF2_GlobalSnapReaderCallbacks_SetParameterUnsignedIntCallback
    c_SetParameterUnsignedIntCallback.argtypes = [ ctypes.POINTER(GlobalSnapReaderCallbacks), _GlobalSnapReaderCallback_FP_ParameterUnsignedInt ]
    c_SetParameterUnsignedIntCallback.restype = ErrorCode
    c_SetParameterUnsignedIntCallback.errcheck = HandleErrorCode
    wrapped_callback = _callback_wrapper(_GlobalSnapReaderCallback_FP_parameterUnsignedIntCallback,
                                         parameterUnsignedIntCallback)
    c_SetParameterUnsignedIntCallback(globalSnapReaderCallbacks, wrapped_callback)

__all__ = [
    'GlobalSnapReaderCallbacks',
    'GlobalSnapReaderCallbacks_New',
    'GlobalSnapReaderCallbacks_Delete',
    'GlobalSnapReaderCallbacks_Clear',
    'GlobalSnapReaderCallbacks_SetUnknownCallback',
    'GlobalSnapReaderCallbacks_SetSnapshotStartCallback',
    'GlobalSnapReaderCallbacks_SetSnapshotEndCallback',
    'GlobalSnapReaderCallbacks_SetMeasurementOnOffCallback',
    'GlobalSnapReaderCallbacks_SetEnterCallback',
    'GlobalSnapReaderCallbacks_SetMpiSendCallback',
    'GlobalSnapReaderCallbacks_SetMpiIsendCallback',
    'GlobalSnapReaderCallbacks_SetMpiIsendCompleteCallback',
    'GlobalSnapReaderCallbacks_SetMpiRecvCallback',
    'GlobalSnapReaderCallbacks_SetMpiIrecvRequestCallback',
    'GlobalSnapReaderCallbacks_SetMpiIrecvCallback',
    'GlobalSnapReaderCallbacks_SetMpiCollectiveBeginCallback',
    'GlobalSnapReaderCallbacks_SetMpiCollectiveEndCallback',
    'GlobalSnapReaderCallbacks_SetOmpForkCallback',
    'GlobalSnapReaderCallbacks_SetOmpAcquireLockCallback',
    'GlobalSnapReaderCallbacks_SetOmpTaskCreateCallback',
    'GlobalSnapReaderCallbacks_SetOmpTaskSwitchCallback',
    'GlobalSnapReaderCallbacks_SetMetricCallback',
    'GlobalSnapReaderCallbacks_SetParameterStringCallback',
    'GlobalSnapReaderCallbacks_SetParameterIntCallback',
    'GlobalSnapReaderCallbacks_SetParameterUnsignedIntCallback',
]




# Generated by OTF2 Template Engine

from collections import defaultdict, OrderedDict
from itertools import chain
from functools import wraps

import _otf2

from . import definitions
from .error import TraceReaderError, TraceWriterError
from .enums import Undefined, GroupType
from .definitions import ClockProperties


def _use_doc_from(doc):
    def decorator(func):
        doc_str = "Returns the {} definition with matching arguments\n".format(doc.__name__)
        if isinstance(doc, definitions._RefDefinition):
            doc_str += "If there is no matching definition, one will be created.\n"
        doc_str += "\n"
        doc_str += doc.__doc__

        func.__doc__ = doc_str
        return func
    return decorator


class _Registry(object):
    def __init__(self, definition_registry, default_type=None):
        self._definition_registry = definition_registry
        self._default_type = default_type
        self._extra_callbacks = {}

    def _callback_gen(self, definition_type=None):
        if definition_type is None:
            definition_type = self._default_type

        def callback(definition_registry, *args):
            definition = definition_type._construct(definition_registry, *args)
            try:
                self._register(definition)
            except KeyError:
                raise TraceReaderError("Duplicate definition: {}".format(definition))
        self._extra_callbacks[definition_type] = callback
        return callback

    def create(self, *args, **kwargs):
        t = kwargs.pop('_type', self._default_type)
        obj = t(self, *args, **kwargs)
        self._register(obj)
        return obj


class _ListRegistry(_Registry):
    def __init__(self, definition_registry, default_type=None):
        self._elements = []
        super(_ListRegistry, self).__init__(definition_registry, default_type)

    def _register(self, obj):
        self._elements.append(obj)

    def __iter__(self):
        """ Do not try to add elements during iteration """
        for elem in self._elements:
            yield elem


class _RefRegistry(_Registry):
    """
    All objects recorded by this registry must have:
    - a ref (id) of type int as first constructor parameter.
      This must not be changed once an event with that ref is written in the trace
    - a name of type str as a second constructor parameter.
      If this is changed, the _elements_by_name must be updated
     """
    def __init__(self, definition_registry, undefined_ref, default_type=None):
        self._elements_by_key = dict()
        self._elements_by_ref = OrderedDict({undefined_ref.value: None})

        self._ref = -1
        self._extra_callbacks = dict()
        super(_RefRegistry, self).__init__(definition_registry, default_type)

    def _next_ref(self):
        self._ref += 1
        return self._ref

    def create(self, *args, **kwargs):
        ref = self._next_ref()
        return super(_RefRegistry, self).create(ref, *args, **kwargs)

    def get(self, *args, **kwargs):
        t = kwargs.pop('_type', self._default_type)
        key = t._new_key(self, *args, **kwargs)
        try:
            return self._elements_by_key[key]
        except KeyError:
            pass
        return self.create(*args, _type=t, **kwargs)

    def _register(self, obj):
        if obj._ref in self._elements_by_ref:
            raise KeyError("Trying to register definition with existing reference.")
        self._elements_by_key[obj._key] = obj
        self._elements_by_ref[obj._ref] = obj
        # Maybe redundant, don't care
        self._ref = max(self._ref, obj._ref)

    def __iter__(self):
        """ Do not try to add elements during iteration """
        for elem in self._elements_by_ref.values():
            # Skip the Undefined entry
            if elem is not None:
                yield elem

    def __getitem__(self, ref):
        return self._elements_by_ref[ref]

    def __contains__(self, item):
        return item._key in self._elements_by_key

    def __len__(self):
        return len(self._elements_by_ref) - 1


class _StringRegistry(_RefRegistry):
    """
    This is a special registry for strings.
    """
    def __init__(self, definition_registry):
        super(_StringRegistry, self).__init__(definition_registry, Undefined.STRING,
                                              definitions.String)

    def __getitem__(self, ref):
        if self._elements_by_ref[ref] is None:
            return None
        return self._elements_by_ref[ref].name

    def get_ref(self, name, writer=None):
        """
        Get a definition object, and write it in the writer if it was not yet created.
        This means, that all previously created definitions should already be written to the writer.
        Returns the ref of the definition for simplicity.
        """
        if not isinstance(name, str):
            raise TypeError(f"trying to get string reference for non-str type {type(name)}")
        key = definitions.String._new_key(self, name)
        try:
            return self._elements_by_key[key]._ref
        except KeyError:
            pass
        obj = self.create(name, _type=definitions.String)
        if writer:
            obj.write(writer, self._definition_registry)
        return obj._ref


class DefinitionRegistry(object):
    """
        This class stores all definitions.
    """
    def __init__(self):
        self.clock_properties = None
        self._strings = _StringRegistry(self)
        @otf2 for def in defs|global_defs|referable_defs|shared_defs|first_born_defs|exclude(String):
        self._@@def.ref_lower|pluralize@@ = _RefRegistry(self, Undefined.@@def.ref_upper@@)
        @otf2 endfor
        @otf2 for def in defs|global_defs|referable_defs|not_shared_defs|exclude(String):
        self._@@def.lower|pluralize@@ = _RefRegistry(self, Undefined.@@def.upper@@, definitions.@@def.name@@)
        @otf2 endfor

        @otf2 for def in defs|global_defs|not_referable_defs|exclude(ClockProperties):
        self._@@def.lower|pluralize@@ = _ListRegistry(self, definitions.@@def.name@@)
        @otf2 endfor

    def write(self, writer):
        # Warning, the order here is important.
        # That is sometimes not trivial: Groups can refer metrics, but metric instances can
        # refer to groups... So we first write the non-metric groups and then, after metrics,
        # the metric groups
        # Note: The split writing metric groups may violate the strong monotonic order of
        # definition reference

        # Note: We do not write metric classes and instances separately. The order given by the
        # reference will always reflect dependencies from instances on classes
        for thing in chain(self._strings,
                           self._paradigms,
                           self._attributes,
                           self._system_tree_nodes,
                           self._system_tree_node_properties,
                           self._system_tree_node_domains,
                           self._location_groups,
                           self._location_group_properties,
                           self._locations,
                           self._location_properties,
                           self._regions,
                           filter(lambda g: g.group_type != GroupType.METRIC, self._groups),
                           self._metric_members,
                           self._metrics,
                           self._metric_class_recorders,
                           filter(lambda g: g.group_type == GroupType.METRIC, self._groups),
                           self._comms,
                           self._parameters,
                           self._callsites,
                           self._callpaths,
                           self._callpath_parameters,
                           self._rma_wins,
                           self._cart_dimensions,
                           self._cart_topologies,
                           self._cart_coordinates,
                           self._source_code_locations,
                           self._calling_contexts,
                           self._calling_context_properties,
                           self._interrupt_generators,
                           self._paradigm_properties,
                           self._io_paradigms,
                           self._io_files,
                           self._io_file_properties,
                           self._io_handles,
                           self._io_pre_created_handle_states):
            thing.write(writer, self)
        # Make sure all location writers are created, otherwise files will be missing
        for location in self._locations:
            writer._archive.event_writer_from_location(location)

    def _set_global_def_reader_callbacks(self, cbs):
        _otf2.GlobalDefReaderCallbacks_SetClockPropertiesCallback(
            cbs, self._set_clock_properties)
        @otf2 for def in defs|global_defs|shared_defs:
        _otf2.GlobalDefReaderCallbacks_Set@@def.name@@Callback(
            cbs, self._@@def.ref_lower|pluralize@@._callback_gen(definitions.@@def.name@@))
        @otf2 endfor
        @otf2 for def in defs|global_defs|not_shared_defs|exclude(ClockProperties):
        _otf2.GlobalDefReaderCallbacks_Set@@def.name@@Callback(
            cbs, self._@@def.lower|pluralize@@._callback_gen())
        @otf2 endfor

    def _set_clock_properties(self, _, *args):
        if self.clock_properties is not None:
            raise TraceReaderError("Duplicate ClockProperties entry.")
        self.clock_properties = ClockProperties(None, *args)

    def _string(self, name):
        """
        This function is used internally to intern String definitions.
        They are not exposed publicly, instead, plain str objects are used
        """
        assert type(name) is str
        return self._strings.get(name)

    def _get(self, def_type, ref):
        return self._registry_for_type(def_type)[ref]

    def _registry_for_type(self, def_type):
        if def_type is str:
            return self._strings
        return getattr(self, def_type._registry_name)

    def metric(self, name, *args, **kwargs):
        """
            A convenience method to create a magic metric.
            Details may change in future.

            :param str name: The name for the metric, required argument
        """
        return self.metric_class([self.metric_member(name, *args, **kwargs)])
    @otf2 for def in defs|global_defs|shared_defs:

    @_use_doc_from(definitions.@@def.name@@)
    def @@def.lower@@(self, *args, **kwargs):
        return self._@@def.ref_lower|pluralize@@.get(*args, _type=definitions.@@def.name@@, **kwargs)
    @otf2 endfor
    @otf2 for def in defs|global_defs|referable_defs|not_shared_defs|exclude(String):

    @_use_doc_from(definitions.@@def.name@@)
    def @@def.lower@@(self, *args, **kwargs):
        return self._@@def.lower|pluralize@@.get(*args, **kwargs)
    @otf2 endfor
    @otf2 for def in defs|global_defs|not_referable_defs|exclude(ClockProperties):

    @_use_doc_from(definitions.@@def.name@@)
    def @@def.lower@@(self, *args, **kwargs):
        return self._@@def.lower|pluralize@@.create(*args, **kwargs)
    @otf2 endfor

    @property
    def strings(self):
        """
            A :py:class:`_Registry` instance representing all String definitons.
        """
        return self._strings
    @otf2 for def in defs|global_defs|shared_defs:

    @property
    def @@def.lower|pluralize@@(self):
        return filter(lambda m: isinstance(m, definitions.@@def.name@@), self._@@def.ref_lower|pluralize@@)
    @otf2 endfor
    @otf2 for def in defs|global_defs|referable_defs|first_born_defs|exclude(String):

    @property
    def @@def.ref_lower|pluralize@@(self):
        """
            A :py:class:`_Registry` instance representing all @@def.ref_name@@ definitons.
        """
        return self._@@def.ref_lower|pluralize@@
    @otf2 endfor
    @otf2 for def in defs|global_defs|not_referable_defs|exclude(ClockProperties):

    @property
    def @@def.lower|pluralize@@(self):
        """
            A :py:class:`_Registry` instance representing all @@def.name@@ definitons.
        """
        return self._@@def.lower|pluralize@@
    @otf2 endfor



#!/usr/bin/env python

import otf2
from otf2.events import *
import time
from tempfile import mkdtemp
import os
import otf2_writer


def read_trace(archive_name="TestArchive/traces.otf2"):
    with otf2.reader.open(archive_name) as trace:
        print("Read {} definitions".format(len(trace.definitions.strings)))

        for location, event in trace.events:
            print("Encountered event on location {} at {}".format(location, event))


if __name__ == '__main__':
    trace_dir = mkdtemp(prefix="trace_test.", dir=os.getcwd())
    os.rmdir(trace_dir)
    otf2_writer.generate_trace(trace_dir)
    print("Created trace in {}".format(trace_dir))
    read_trace("{}/traces.otf2".format(trace_dir))
    print("Read trace from {}/traces.otf2".format(trace_dir))




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import EnumBase, EnumBitset

@otf2 with v = version
VERSION_MAJOR = @@v.major@@
VERSION_MINOR = @@v.minor@@
VERSION_BUGFIX = @@v.bugfix@@
VERSION_SUFFIX = "@@v.suffix@@"
VERSION = "@@v@@"
@otf2 endwith

# Symbols for undefined values

UNDEFINED_UINT8 = ctypes.c_uint8(-1)
UNDEFINED_INT8 = ctypes.c_int8(0x80)
UNDEFINED_UINT16 = ctypes.c_uint16(-1)
UNDEFINED_INT16 = ctypes.c_int16(0x8000)
UNDEFINED_UINT32 = ctypes.c_uint32(-1)
UNDEFINED_INT32 = ctypes.c_int32(0x80000000)
UNDEFINED_UINT64 = ctypes.c_uint64(-1)
UNDEFINED_INT64 = ctypes.c_int64(0x8000000000000000)

UNDEFINED_TYPE = UNDEFINED_UINT8

# Handle Typedefs

TimeStamp = ctypes.c_uint64
UNDEFINED_TIMESTAMP = UNDEFINED_UINT64

@otf2 for def in defs|referable_defs|first_born_defs:
@@def.self_attr.type.py_ctype@@ = @@def.self_attr.type.py_underlying_ctype@@
UNDEFINED_@@def.ref_upper@@ = @@def.self_attr.type.py_ctype@@(UNDEFINED_@@def.self_attr.type.otf2|upper@@.value)

@otf2 endfor
CHUNK_SIZE_MIN = ctypes.c_uint64(256 * 1024)

CHUNK_SIZE_MAX = ctypes.c_uint64(1024 * 1024 * 16)

class FileMode(EnumBase, ctypes.c_uint8):
    pass

FILEMODE_WRITE = FileMode._construct(0, "WRITE")
FILEMODE_READ = FileMode._construct(1, "READ")
FILEMODE_MODIFY = FileMode._construct(2, "MODIFY")

class Compression(EnumBase, ctypes.c_uint8):
    pass

COMPRESSION_UNDEFINED = Compression._construct(0, "UNDEFINED")
COMPRESSION_NONE = Compression._construct(1, "NONE")
COMPRESSION_ZLIB = Compression._construct(2, "ZLIB")

# CallbackCode is even an enum on the C-Side
# usually used as return value, evtl. should be a restype method?
class CallbackCode(EnumBase, ctypes.c_int):
    pass

CALLBACK_SUCCESS = CallbackCode._construct(0, "SUCCESS")
CALLBACK_INTERRUPT = CallbackCode._construct(1, "INTERRUPT")
CALLBACK_ERROR = CallbackCode._construct(1, "ERROR")

class FlushType(EnumBase, ctypes.c_uint8):
    pass

NO_FLUSH = FlushType._construct(0, "NO_FLUSH")
FLUSH = FlushType._construct(1, "FLUSH")

class Hint(EnumBase, ctypes.c_uint8):
    pass

HINT_GLOBAL_READER = Hint._construct(0, "GLOBAL_READER")

@otf2 for enum in enums|general_enums:
@otf2  if enum is bitset_enum:
class @@enum.type.py_ctype@@(EnumBitset, @@enum.type.py_underlying_ctype@@):
@otf2  else:
class @@enum.type.py_ctype@@(EnumBase, @@enum.type.py_underlying_ctype@@):
@otf2  endif
    pass

@otf2  for entry in enum.entries|py_def_enum_entries:
@@entry.name[5:]@@ = @@enum.type.py_ctype@@._construct(@@entry.value@@, "@@entry.suffix@@")
@otf2 for alias in entry.aliases:
@@alias.name[5:]@@ = @@enum.type.py_ctype@@._construct(@@entry.value@@, "@@alias.suffix@@")
@otf2 endfor
@otf2  endfor
@otf2  if enum is enum_with_terminator:
@@enum.terminator.name[5:]@@ = @@enum.type.py_ctype@@(@@enum.terminator.value@@)
@otf2  endif

@otf2 endfor
//@ This is a special case because TRUE is LITERALLY defined as '!OTF2_FALSE'
TRUE = Boolean._construct(1, "TRUE")

__all__ = [
    'VERSION_MAJOR',
    'VERSION_MINOR',
    'VERSION_BUGFIX',
    'VERSION_SUFFIX',
    'VERSION',
    'UNDEFINED_UINT8',
    'UNDEFINED_INT8',
    'UNDEFINED_UINT16',
    'UNDEFINED_INT16',
    'UNDEFINED_UINT32',
    'UNDEFINED_INT32',
    'UNDEFINED_UINT64',
    'UNDEFINED_INT64',
    'UNDEFINED_TYPE',
    'TimeStamp',
    'UNDEFINED_TIMESTAMP',
    @otf2 for def in defs|referable_defs|first_born_defs
    '@@def.self_attr.type.py_ctype@@',
    'UNDEFINED_@@def.ref_upper@@',
    @otf2 endfor
    'CHUNK_SIZE_MIN',
    'CHUNK_SIZE_MAX',
    'FileMode',
    'FILEMODE_WRITE',
    'FILEMODE_READ',
    'FILEMODE_MODIFY',
    'Compression',
    'COMPRESSION_UNDEFINED',
    'COMPRESSION_NONE',
    'COMPRESSION_ZLIB',
    'CallbackCode',
    'CALLBACK_SUCCESS',
    'CALLBACK_INTERRUPT',
    'CALLBACK_ERROR',
    'FlushType',
    'NO_FLUSH',
    'FLUSH',
    @otf2 for enum in enums|general_enums:
    '@@enum.type.py_ctype@@',
    @otf2 for entry in enum.entries:
    '@@entry.name[5:]@@',
    @otf2 endfor
    @otf2 if enum is enum_with_terminator:
    '@@enum.terminator.name[5:]@@',
    @otf2 endif
    @otf2 endfor
    'Hint',
    'HINT_GLOBAL_READER'
]



#!/usr/bin/env python

import otf2
from otf2.enums import Type
import time
from tempfile import mkdtemp
import os

TIMER_GRANULARITY = 1000000


def t():
    return int(round(time.time() * TIMER_GRANULARITY))


def generate_trace(archive_name="TestArchive", num_locations=1):
    with otf2.writer.open(archive_name, timer_resolution=TIMER_GRANULARITY) as trace:

        function = trace.definitions.region("My Function")

        parent_node = trace.definitions.system_tree_node("root node")
        system_tree_node = trace.definitions.system_tree_node("myHost", parent=parent_node)
        trace.definitions.system_tree_node_property(system_tree_node, "color", value="black")
        trace.definitions.system_tree_node_property(system_tree_node, "rack #", value=42)

        location_group = trace.definitions.location_group("Initial Process",
                                                          system_tree_parent=system_tree_node)

        attr = trace.definitions.attribute("StringTest", "A test attribute", Type.STRING)
        float_attr = trace.definitions.attribute("FloatTest", "Another test attribute",
                                                 Type.DOUBLE)

        # Get convenience metric object and write one metric event
        temperature = trace.definitions.metric("Time since last coffee", unit="min")

        # Get metric members
        temp_member = trace.definitions.metric_member("Instructions", "C", otf2.MetricType.OTHER,
                                                      otf2.MetricMode.ABSOLUTE_POINT)
        power_member = trace.definitions.metric_member("Mem", "W")

        # Add metric members to the metric class object
        mclass = trace.definitions.metric_class([temp_member, power_member])

        for location in range(num_locations):
            writer = trace.event_writer("Main Thread {}".format(location), group=location_group)

            # Write enter and leave event
            writer.enter(t(), function, {attr: "Hello World"})

            writer.metric(t(), temperature, 72.0)

            # Add metric object to the location object
            writer.metric(t(), mclass, [42.0, 12345.6])

            writer.leave(t(), function, attributes={float_attr: 42.0, attr: "Wurst?"})

        trace.definitions.group("PROCESSXTHREADS_GROUP_LOCATIONS",
                                otf2.GroupType.COMM_LOCATIONS,
                                otf2.Paradigm.MEASUREMENT_SYSTEM,
                                otf2.GroupFlag.NONE,
                                list(trace.definitions.locations))
        group = trace.definitions.group("PROXESSXTHREADS_GROUP",
                                        otf2.GroupType.COMM_GROUP,
                                        otf2.Paradigm.MEASUREMENT_SYSTEM,
                                        otf2.GroupFlag.NONE,
                                        list(range(num_locations)))
        dimensions = trace.definitions.cart_dimension("Process", num_locations),\
                     trace.definitions.cart_dimension("Thread", 1)
        comm = trace.definitions.comm("Process x Threads CPU Locations", group)
        topo = trace.definitions.cart_topology("Process x Thread", comm, dimensions)
        for rank in range(len(trace.definitions.locations)):
            trace.definitions.cart_coordinate(topo, rank, (rank, 0))

        return trace.definitions


if __name__ == '__main__':
    trace_dir = mkdtemp(prefix="trace_test.", dir=os.getcwd())
    os.rmdir(trace_dir)
    generate_trace(trace_dir)
    print("Created trace in {}".format(trace_dir))



import os
import os.path
import unittest
from shutil import rmtree
from tempfile import mkdtemp
import time

import otf2

from otf2_writer import generate_trace

TIMER_GRANULARITY = 1000000


def t():
    return int(round(time.time() * TIMER_GRANULARITY))


class TestOTF2UserData(unittest.TestCase):

    archive_name = "otf2_test_trace"

    def setUp(self):
        self.old_cwd = os.getcwd()
        self.tmp_dirname = mkdtemp(prefix=os.path.basename(os.path.abspath(__file__))[:-3] + '_tmp', dir=self.old_cwd)
        os.chdir(self.tmp_dirname)

    def tearDown(self):
        os.chdir(self.old_cwd)
        if os.getenv('KEEP_TEST_OUTPUT', '') != '':
            print(self.tmp_dirname)
        else:
            rmtree(self.tmp_dirname)

    def read_trace(self, archive_name):
        count_selected = 0
        with otf2.reader.open(archive_name) as trace:
            selected_location = trace.definitions.locations[1]
            for location, event in trace.events(selected_location):
                self.assertEqual(location, selected_location)
                count_selected += 1

            with self.assertRaises(RuntimeError):
                trace.events(selected_location)

        count_reference = 0
        with otf2.reader.open(archive_name) as trace:
            selected_location = trace.definitions.locations[1]
            for location, event in trace.events:
                if location == selected_location:
                    count_reference += 1

            with self.assertRaises(EOFError):
                for _ in trace.events:
                    pass

        self.assertEqual(count_selected, count_reference)

    def test_integrity(self):
        self.definitions = generate_trace(self.tmp_dirname, 2)
        self.read_trace(os.path.join(self.tmp_dirname, "traces.otf2"))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# This file is part of the Score-P software (http://www.score-p.org)
#
# Copyright (c) 2015-2016, 2021,
# Technische Universitaet Dresden, Germany
#
# This software may be modified and distributed under the terms of
# a BSD-style license.  See the COPYING file in the package base
# directory for details.
#

import os
from tempfile import mkdtemp

import _otf2


def time_stamp():
    time_stamp.counter += 1
    return time_stamp.counter


time_stamp.counter = 0


def pre_flush(user_data, file_type, location, caller_data, final):
    return _otf2.FLUSH


def post_flush(user_data, file_type, location):
    return time_stamp()


def write_archive(path, name):
    archive = _otf2.Archive_Open(path, name,
                                 _otf2.FILEMODE_WRITE,
                                 1024 * 1024, 4 * 1024 * 1024,
                                 _otf2.SUBSTRATE_POSIX, _otf2.COMPRESSION_NONE)

    flush_callbacks = _otf2.FlushCallbacks(pre_flush=pre_flush, post_flush=post_flush)
    _otf2.Archive_SetFlushCallbacks(archive, flush_callbacks, None)
    _otf2.Archive_SetSerialCollectiveCallbacks(archive)

    write_events(archive)
    write_definitions(archive)
    write_global_definitions(archive)

    _otf2.Archive_Close(archive)


def write_definitions(archive):
    def_writer = _otf2.Archive_GetDefWriter(archive, 0)


def write_global_definitions(archive):
    global_def_writer = _otf2.Archive_GetGlobalDefWriter(archive)

    _otf2.GlobalDefWriter_WriteClockProperties(global_def_writer, 1, 0, 2, _otf2.UNDEFINED_TIMESTAMP)

    # GlobalDefWriter, StringRef, const char*
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 0, "")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 1, "Initial Process")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 2, "Main Thread")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 3, "My Function")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 4, "other function")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 5, "compute something")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 6, "myHost")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 7, "node")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 8, "Temperature")
    _otf2.GlobalDefWriter_WriteString(global_def_writer, 9, "C")

    # GlobalDefWriter, RegionRef, StringRef, StringRef, StringRef, RegionRole,
    # Paradigm, RegionFlag, StringRef, uint32_t, uint32_t
    _otf2.GlobalDefWriter_WriteRegion(global_def_writer, 0, 3, 4, 5, _otf2.REGION_ROLE_FUNCTION,
                                      _otf2.PARADIGM_USER, _otf2.REGION_FLAG_NONE, 0, 0, 0)
    # GlobalDefWriter, SystemTreeNodeRef, StringRef, StringRef, SystemTreeNodeRef
    _otf2.GlobalDefWriter_WriteSystemTreeNode(global_def_writer, 0, 6, 7, _otf2.UNDEFINED_SYSTEM_TREE_NODE)
    # GlobalDefWriter, LocationGroupRef, StringRef, LocationGroupType, SystemTreeNodeRef, LocationGroupRef
    _otf2.GlobalDefWriter_WriteLocationGroup(global_def_writer, 0, 1, _otf2.LOCATION_GROUP_TYPE_PROCESS, 0,
                                             _otf2.UNDEFINED_LOCATION_GROUP)
    # GlobalDefWriter, LocationRef, StringRef, LocationType, uint64_t, LocationGroupRef
    _otf2.GlobalDefWriter_WriteLocation(global_def_writer, 0, 2, _otf2.LOCATION_TYPE_CPU_THREAD, 2, 0)

    write_paradigm(global_def_writer)


def write_events(archive):
    _otf2.Archive_OpenEvtFiles(archive)

    evt_writer = _otf2.Archive_GetEvtWriter(archive, 0)

    _otf2.EvtWriter_Enter(evt_writer, None, time_stamp(), 0)
    _otf2.EvtWriter_Leave(evt_writer, None, time_stamp(), 0)

    _otf2.Archive_CloseEvtWriter(archive, evt_writer)
    _otf2.Archive_CloseEvtFiles(archive)


def write_paradigm(global_def_writer):
    str_id_mpi = write_string(global_def_writer, "MPI")
    _otf2.GlobalDefWriter_WriteParadigm(global_def_writer,
                                        _otf2.PARADIGM_MPI,
                                        str_id_mpi,
                                        _otf2.PARADIGM_CLASS_PROCESS)

    attribute_value = _otf2.AttributeValue()
    attribute_value.stringRef = write_string(global_def_writer, "MPI_COMM_${id}")
    _otf2.GlobalDefWriter_WriteParadigmProperty(global_def_writer,
                                                _otf2.PARADIGM_MPI,
                                                _otf2.PARADIGM_PROPERTY_COMM_NAME_TEMPLATE,
                                                _otf2.TYPE_STRING,
                                                attribute_value)

    attribute_value.stringRef = write_string(global_def_writer, "MPI_WIN_${id}")
    _otf2.GlobalDefWriter_WriteParadigmProperty(global_def_writer,
                                                _otf2.PARADIGM_MPI,
                                                _otf2.PARADIGM_PROPERTY_RMA_WIN_NAME_TEMPLATE,
                                                _otf2.TYPE_STRING,
                                                attribute_value)


#    type_ =_otf2.
#    _otf2.AttributeValue_SetBoolean(_otf2.FALSE, type_, attribute_value)
#    _otf2.GlobalDefWriter_WriteParadigmProperty(global_def_writer,
#                                                _otf2.PARADIGM_OPENMP,
#                                                _otf2.PARADIGM_PROPERTY_RMA_ONLY,
#                                                type_,
#                                                attribute_value)

def write_string(global_def_writer, string):
    _otf2.GlobalDefWriter_WriteString(global_def_writer, write_string.id, string)
    write_string.id += 1
    return write_string.id - 1


write_string.id = 0

if __name__ == '__main__':
    trace_dir = mkdtemp(prefix="trace_test.", dir=os.getcwd())
    os.rmdir(trace_dir)
    write_archive(trace_dir, "traces")




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .AttributeList import AttributeList
from .Definitions import *
from .Events import *


class SnapWriter(ctypes.Structure):
    pass

def SnapWriter_GetLocationID(writer, locationID = LocationRef()):
    c_GetLocationID = conf.lib.OTF2_SnapWriter_GetLocationID
    c_GetLocationID.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(LocationRef) ]
    c_GetLocationID.restype = ErrorCode
    c_GetLocationID.errcheck = HandleErrorCode
    c_GetLocationID(writer, ctypes.byref(locationID))
    return locationID.value

def SnapWriter_SnapshotStart(writerHandle, attributeList, snapTime, numberOfRecords):
    c_SnapshotStart = conf.lib.OTF2_SnapWriter_SnapshotStart
    c_SnapshotStart.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64 ]
    c_SnapshotStart.restype = ErrorCode
    c_SnapshotStart.errcheck = HandleErrorCode
    c_SnapshotStart(writerHandle, attributeList, snapTime, numberOfRecords)

def SnapWriter_SnapshotEnd(writerHandle, attributeList, snapTime, contReadPos):
    c_SnapshotEnd = conf.lib.OTF2_SnapWriter_SnapshotEnd
    c_SnapshotEnd.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, ctypes.c_uint64 ]
    c_SnapshotEnd.restype = ErrorCode
    c_SnapshotEnd.errcheck = HandleErrorCode
    c_SnapshotEnd(writerHandle, attributeList, snapTime, contReadPos)

def SnapWriter_MeasurementOnOff(writerHandle, attributeList, snapTime, origEventTime, measurementMode):
    c_MeasurementOnOff = conf.lib.OTF2_SnapWriter_MeasurementOnOff
    c_MeasurementOnOff.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, MeasurementMode ]
    c_MeasurementOnOff.restype = ErrorCode
    c_MeasurementOnOff.errcheck = HandleErrorCode
    c_MeasurementOnOff(writerHandle, attributeList, snapTime, origEventTime, measurementMode)

def SnapWriter_Enter(writerHandle, attributeList, snapTime, origEventTime, region):
    c_Enter = conf.lib.OTF2_SnapWriter_Enter
    c_Enter.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, RegionRef ]
    c_Enter.restype = ErrorCode
    c_Enter.errcheck = HandleErrorCode
    c_Enter(writerHandle, attributeList, snapTime, origEventTime, region)

def SnapWriter_MpiSend(writerHandle, attributeList, snapTime, origEventTime, receiver, communicator, msgTag, msgLength):
    c_MpiSend = conf.lib.OTF2_SnapWriter_MpiSend
    c_MpiSend.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64 ]
    c_MpiSend.restype = ErrorCode
    c_MpiSend.errcheck = HandleErrorCode
    c_MpiSend(writerHandle, attributeList, snapTime, origEventTime, receiver, communicator, msgTag, msgLength)

def SnapWriter_MpiIsend(writerHandle, attributeList, snapTime, origEventTime, receiver, communicator, msgTag, msgLength, requestID):
    c_MpiIsend = conf.lib.OTF2_SnapWriter_MpiIsend
    c_MpiIsend.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64 ]
    c_MpiIsend.restype = ErrorCode
    c_MpiIsend.errcheck = HandleErrorCode
    c_MpiIsend(writerHandle, attributeList, snapTime, origEventTime, receiver, communicator, msgTag, msgLength, requestID)

def SnapWriter_MpiIsendComplete(writerHandle, attributeList, snapTime, origEventTime, requestID):
    c_MpiIsendComplete = conf.lib.OTF2_SnapWriter_MpiIsendComplete
    c_MpiIsendComplete.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint64 ]
    c_MpiIsendComplete.restype = ErrorCode
    c_MpiIsendComplete.errcheck = HandleErrorCode
    c_MpiIsendComplete(writerHandle, attributeList, snapTime, origEventTime, requestID)

def SnapWriter_MpiRecv(writerHandle, attributeList, snapTime, origEventTime, sender, communicator, msgTag, msgLength):
    c_MpiRecv = conf.lib.OTF2_SnapWriter_MpiRecv
    c_MpiRecv.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64 ]
    c_MpiRecv.restype = ErrorCode
    c_MpiRecv.errcheck = HandleErrorCode
    c_MpiRecv(writerHandle, attributeList, snapTime, origEventTime, sender, communicator, msgTag, msgLength)

def SnapWriter_MpiIrecvRequest(writerHandle, attributeList, snapTime, origEventTime, requestID):
    c_MpiIrecvRequest = conf.lib.OTF2_SnapWriter_MpiIrecvRequest
    c_MpiIrecvRequest.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint64 ]
    c_MpiIrecvRequest.restype = ErrorCode
    c_MpiIrecvRequest.errcheck = HandleErrorCode
    c_MpiIrecvRequest(writerHandle, attributeList, snapTime, origEventTime, requestID)

def SnapWriter_MpiIrecv(writerHandle, attributeList, snapTime, origEventTime, sender, communicator, msgTag, msgLength, requestID):
    c_MpiIrecv = conf.lib.OTF2_SnapWriter_MpiIrecv
    c_MpiIrecv.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64 ]
    c_MpiIrecv.restype = ErrorCode
    c_MpiIrecv.errcheck = HandleErrorCode
    c_MpiIrecv(writerHandle, attributeList, snapTime, origEventTime, sender, communicator, msgTag, msgLength, requestID)

def SnapWriter_MpiCollectiveBegin(writerHandle, attributeList, snapTime, origEventTime):
    c_MpiCollectiveBegin = conf.lib.OTF2_SnapWriter_MpiCollectiveBegin
    c_MpiCollectiveBegin.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp ]
    c_MpiCollectiveBegin.restype = ErrorCode
    c_MpiCollectiveBegin.errcheck = HandleErrorCode
    c_MpiCollectiveBegin(writerHandle, attributeList, snapTime, origEventTime)

def SnapWriter_MpiCollectiveEnd(writerHandle, attributeList, snapTime, origEventTime, collectiveOp, communicator, root, sizeSent, sizeReceived):
    c_MpiCollectiveEnd = conf.lib.OTF2_SnapWriter_MpiCollectiveEnd
    c_MpiCollectiveEnd.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, CollectiveOp, CommRef, ctypes.c_uint32, ctypes.c_uint64, ctypes.c_uint64 ]
    c_MpiCollectiveEnd.restype = ErrorCode
    c_MpiCollectiveEnd.errcheck = HandleErrorCode
    c_MpiCollectiveEnd(writerHandle, attributeList, snapTime, origEventTime, collectiveOp, communicator, root, sizeSent, sizeReceived)

def SnapWriter_OmpFork(writerHandle, attributeList, snapTime, origEventTime, numberOfRequestedThreads):
    c_OmpFork = conf.lib.OTF2_SnapWriter_OmpFork
    c_OmpFork.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32 ]
    c_OmpFork.restype = ErrorCode
    c_OmpFork.errcheck = HandleErrorCode
    c_OmpFork(writerHandle, attributeList, snapTime, origEventTime, numberOfRequestedThreads)

def SnapWriter_OmpAcquireLock(writerHandle, attributeList, snapTime, origEventTime, lockID, acquisitionOrder):
    c_OmpAcquireLock = conf.lib.OTF2_SnapWriter_OmpAcquireLock
    c_OmpAcquireLock.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint32, ctypes.c_uint32 ]
    c_OmpAcquireLock.restype = ErrorCode
    c_OmpAcquireLock.errcheck = HandleErrorCode
    c_OmpAcquireLock(writerHandle, attributeList, snapTime, origEventTime, lockID, acquisitionOrder)

def SnapWriter_OmpTaskCreate(writerHandle, attributeList, snapTime, origEventTime, taskID):
    c_OmpTaskCreate = conf.lib.OTF2_SnapWriter_OmpTaskCreate
    c_OmpTaskCreate.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint64 ]
    c_OmpTaskCreate.restype = ErrorCode
    c_OmpTaskCreate.errcheck = HandleErrorCode
    c_OmpTaskCreate(writerHandle, attributeList, snapTime, origEventTime, taskID)

def SnapWriter_OmpTaskSwitch(writerHandle, attributeList, snapTime, origEventTime, taskID):
    c_OmpTaskSwitch = conf.lib.OTF2_SnapWriter_OmpTaskSwitch
    c_OmpTaskSwitch.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ctypes.c_uint64 ]
    c_OmpTaskSwitch.restype = ErrorCode
    c_OmpTaskSwitch.errcheck = HandleErrorCode
    c_OmpTaskSwitch(writerHandle, attributeList, snapTime, origEventTime, taskID)

def SnapWriter_Metric(writerHandle, attributeList, snapTime, origEventTime, metric, typeIDs, metricValues):
    numberOfMetrics = len(metricValues)
    assert(numberOfMetrics == len(typeIDs))

    TypeArrayType = Type * numberOfMetrics
    type_ids_list = TypeArrayType()
    type_ids_list[:] = typeIDs

    MetricValueArrayType = MetricValue * numberOfMetrics
    metric_values_list = MetricValueArrayType()
    metric_values_list[:] = metricValues

    c_Metric = conf.lib.OTF2_SnapWriter_Metric
    c_Metric.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp , TimeStamp, MetricRef, ctypes.c_uint8, TypeArrayType, MetricValueArrayType ]
    c_Metric.restype = ErrorCode
    c_Metric.errcheck = HandleErrorCode
    c_Metric(writerHandle, attributeList, snapTime , origEventTime, metric, numberOfMetrics, type_ids_list, metric_values_list)

def SnapWriter_ParameterString(writerHandle, attributeList, snapTime, origEventTime, parameter, string):
    c_ParameterString = conf.lib.OTF2_SnapWriter_ParameterString
    c_ParameterString.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ParameterRef, StringRef ]
    c_ParameterString.restype = ErrorCode
    c_ParameterString.errcheck = HandleErrorCode
    c_ParameterString(writerHandle, attributeList, snapTime, origEventTime, parameter, string)

def SnapWriter_ParameterInt(writerHandle, attributeList, snapTime, origEventTime, parameter, value):
    c_ParameterInt = conf.lib.OTF2_SnapWriter_ParameterInt
    c_ParameterInt.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ParameterRef, ctypes.c_int64 ]
    c_ParameterInt.restype = ErrorCode
    c_ParameterInt.errcheck = HandleErrorCode
    c_ParameterInt(writerHandle, attributeList, snapTime, origEventTime, parameter, value)

def SnapWriter_ParameterUnsignedInt(writerHandle, attributeList, snapTime, origEventTime, parameter, value):
    c_ParameterUnsignedInt = conf.lib.OTF2_SnapWriter_ParameterUnsignedInt
    c_ParameterUnsignedInt.argtypes = [ ctypes.POINTER(SnapWriter), ctypes.POINTER(AttributeList), TimeStamp, TimeStamp, ParameterRef, ctypes.c_uint64 ]
    c_ParameterUnsignedInt.restype = ErrorCode
    c_ParameterUnsignedInt.errcheck = HandleErrorCode
    c_ParameterUnsignedInt(writerHandle, attributeList, snapTime, origEventTime, parameter, value)

__all__ = [
    'SnapWriter',
    'SnapWriter_GetLocationID',
    'SnapWriter_SnapshotStart',
    'SnapWriter_SnapshotEnd',
    'SnapWriter_MeasurementOnOff',
    'SnapWriter_Enter',
    'SnapWriter_MpiSend',
    'SnapWriter_MpiIsend',
    'SnapWriter_MpiIsendComplete',
    'SnapWriter_MpiRecv',
    'SnapWriter_MpiIrecvRequest',
    'SnapWriter_MpiIrecv',
    'SnapWriter_MpiCollectiveBegin',
    'SnapWriter_MpiCollectiveEnd',
    'SnapWriter_OmpFork',
    'SnapWriter_OmpAcquireLock',
    'SnapWriter_OmpTaskCreate',
    'SnapWriter_OmpTaskSwitch',
    'SnapWriter_Metric',
    'SnapWriter_ParameterString',
    'SnapWriter_ParameterInt',
    'SnapWriter_ParameterUnsignedInt',
]




''' Generated by OTF2 Template Engine '''

import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import *
from .Definitions import *
from .AttributeValue import AttributeValue


class GlobalDefWriter(ctypes.Structure):
    pass

def GlobalDefWriter_GetNumberOfDefinitions(writerHandle, numberOfDefinitions = ctypes.c_uint64()):
    c_GetNumberOfDefinitions = conf.lib.OTF2_GlobalDefWriter_GetNumberOfDefinitions
    c_GetNumberOfDefinitions.argtypes = [ ctypes.POINTER(GlobalDefWriter), ctypes.POINTER(ctypes.c_uint64) ]
    c_GetNumberOfDefinitions.restype = ErrorCode
    c_GetNumberOfDefinitions.errcheck = HandleErrorCode
    c_GetNumberOfDefinitions(writerHandle, ctypes.byref(numberOfDefinitions))
    return numberOfDefinitions.value

def GlobalDefWriter_GetNumberOfLocations(writerHandle, numberOfLocations = ctypes.c_uint64()):
    c_GetNumberOfLocations = conf.lib.OTF2_GlobalDefWriter_GetNumberOfLocations
    c_GetNumberOfLocations.argtypes = [ ctypes.POINTER(GlobalDefWriter), ctypes.POINTER(ctypes.c_uint64) ]
    c_GetNumberOfLocations.restype = ErrorCode
    c_GetNumberOfLocations.errcheck = HandleErrorCode
    c_GetNumberOfLocations(writerHandle, ctypes.byref(numberOfLocations))
    return numberOfLocations.value

@otf2 for def in defs:
@otf2  if def == IoParadigm:
def GlobalDefWriter_WriteIoParadigm(writerHandle, self, identification, name, ioParadigmClass, ioParadigmFlags, properties, types, values):
    numberOfProperties = len(properties)
    assert numberOfProperties == len(types) == len(values)
    properties_array = (IoParadigmProperty * numberOfProperties)()
    properties_array[:] = properties
    types_array = (Type * numberOfProperties)()
    types_array[:] = types
    values_array = (AttributeValue * numberOfProperties)()
    values_array[:] = values

    c_WriteIoParadigm = conf.lib.OTF2_GlobalDefWriter_WriteIoParadigm
    c_WriteIoParadigm.argtypes = [ ctypes.POINTER(GlobalDefWriter), IoParadigmRef, StringRef, StringRef, IoParadigmClass, IoParadigmFlag, ctypes.c_uint8, ctypes.POINTER(IoParadigmProperty), ctypes.POINTER(Type), ctypes.POINTER(AttributeValue) ]
    c_WriteIoParadigm.restype = ErrorCode
    c_WriteIoParadigm.errcheck = HandleErrorCode
    c_WriteIoParadigm(writerHandle, self, identification, name, ioParadigmClass, ioParadigmFlags,
                      numberOfProperties, properties_array, types_array, values_array)

@otf2  elif def == Group:
def GlobalDefWriter_WriteGroup(writerHandle, self, name, groupType, paradigm, groupFlags, members):
    numberOfMembers = len(members)
    array_type = ctypes.c_uint64 * numberOfMembers
    members_array = array_type()
    members_array[:] = members

    c_WriteGroup = conf.lib.OTF2_GlobalDefWriter_WriteGroup
    c_WriteGroup.argtypes = [ ctypes.POINTER(GlobalDefWriter), GroupRef, StringRef, GroupType, Paradigm, GroupFlag,
                              ctypes.c_uint32, array_type ]
    c_WriteGroup.restype = ErrorCode
    c_WriteGroup.errcheck = HandleErrorCode
    c_WriteGroup(writerHandle, self, name, groupType, paradigm, groupFlags, numberOfMembers, members_array)

@otf2  elif def == MetricClass:
def GlobalDefWriter_WriteMetricClass(writerHandle, self, metricMembers, metricOccurrence, recorderKind):
    numberOfMetrics = len(metricMembers)
    array_type = MetricMemberRef * numberOfMetrics
    metric_member_array = array_type()
    metric_member_array[:] = metricMembers

    c_WriteMetricClass = conf.lib.OTF2_GlobalDefWriter_WriteMetricClass
    c_WriteMetricClass.argtypes = [ ctypes.POINTER(GlobalDefWriter), MetricRef, ctypes.c_uint8, array_type,
                                    MetricOccurrence, RecorderKind ]
    c_WriteMetricClass.restype = ErrorCode
    c_WriteMetricClass.errcheck = HandleErrorCode
    c_WriteMetricClass(writerHandle, MetricRef(self), numberOfMetrics, metric_member_array,
                       metricOccurrence, recorderKind)

@otf2  elif def == CartTopology:
def GlobalDefWriter_WriteCartTopology(writerHandle, self, name, communicator, cartDimensions):
    numberOfDimensions = len(cartDimensions)
    array_type = CartDimensionRef * numberOfDimensions
    cart_dimensions_array = array_type()
    cart_dimensions_array[:] = cartDimensions

    c_WriteCartTopology = conf.lib.OTF2_GlobalDefWriter_WriteCartTopology
    c_WriteCartTopology.argtypes = [ ctypes.POINTER(GlobalDefWriter), CartTopologyRef, StringRef, CommRef,
                                     ctypes.c_uint8, array_type ]
    c_WriteCartTopology.restype = ErrorCode
    c_WriteCartTopology.errcheck = HandleErrorCode
    c_WriteCartTopology(writerHandle, self, name, communicator, numberOfDimensions, cart_dimensions_array)

@otf2  elif def == CartCoordinate:
def GlobalDefWriter_WriteCartCoordinate(writerHandle, cartTopology, rank, coordinates):
    numberOfDimensions = len(coordinates)
    array_type = ctypes.c_uint32 * numberOfDimensions
    coordinates_array = array_type()
    coordinates_array[:] = coordinates

    c_WriteCartCoordinate = conf.lib.OTF2_GlobalDefWriter_WriteCartCoordinate
    c_WriteCartCoordinate.argtypes = [ ctypes.POINTER(GlobalDefWriter), CartTopologyRef, ctypes.c_uint32,
                                       ctypes.c_uint8, array_type ]
    c_WriteCartCoordinate.restype = ErrorCode
    c_WriteCartCoordinate.errcheck = HandleErrorCode
    c_WriteCartCoordinate(writerHandle, cartTopology, rank, numberOfDimensions, coordinates_array)

@otf2  else:
def GlobalDefWriter_Write@@def.name@@(writerHandle@@def.py_funcargs()@@):
    c_Write@@def.name@@ = conf.lib.OTF2_GlobalDefWriter_Write@@def.name@@
    c_Write@@def.name@@.argtypes = [ ctypes.POINTER(GlobalDefWriter)@@def.py_funcargtypes()@@ ]
    c_Write@@def.name@@.restype = ErrorCode
    c_Write@@def.name@@.errcheck = HandleErrorCode
    c_Write@@def.name@@(writerHandle@@def.py_callargs(pass_by_ref=True)@@)

@otf2  endif
@otf2 endfor
__all__ = [
    'GlobalDefWriter',
    'GlobalDefWriter_GetNumberOfDefinitions',
    'GlobalDefWriter_GetNumberOfLocations',
    @otf2 for def in defs|global_defs:
    'GlobalDefWriter_Write@@def.name@@',
    @otf2 endfor
]



import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .GeneralDefinitions import TimeStamp
from .Marker import MarkerRef, MarkerScope, MarkerSeverity


class MarkerWriter(ctypes.Structure):
    pass

def MarkerWriter_WriteDefMarker(writerHandle, self, markerGroup, markerCategory, severity):
    c_WriteDefMarker = conf.lib.OTF2_MarkerWriter_WriteDefMarker
    c_WriteDefMarker.argtypes = [ ctypes.POINTER(MarkerWriter), MarkerRef, StrParam, StrParam, MarkerSeverity ]
    c_WriteDefMarker.restype = ErrorCode
    c_WriteDefMarker.errcheck = HandleErrorCode
    c_WriteDefMarker(writerHandle, self, markerGroup, markerCategory, severity)

def MarkerWriter_WriteMarker(writerHandle, timestamp, duration, marker, scope, scopeRef, text):
    c_WriteMarker = conf.lib.OTF2_MarkerWriter_WriteMarker
    c_WriteMarker.argtypes = [ ctypes.POINTER(MarkerWriter), TimeStamp, TimeStamp, MarkerRef, MarkerScope, ctypes.c_uint64, StrParam ]
    c_WriteMarker.restype = ErrorCode
    c_WriteMarker.errcheck = HandleErrorCode
    c_WriteMarker(writerHandle, timestamp, duration, marker, scope, scopeRef, text)

__all__ = [
    'MarkerWriter',
    'MarkerWriter_WriteDefMarker',
    'MarkerWriter_WriteMarker'
]



import ctypes

from .Config import conf, StrParam
from .ErrorCodes import ErrorCode, HandleErrorCode
from .MarkerReaderCallbacks import MarkerReaderCallbacks


class MarkerReader(ctypes.Structure):
    pass

def MarkerReader_ReadMarkers(reader, recordsToRead):
    c_ReadMarkers = conf.lib.OTF2_MarkerReader_ReadMarkers
    c_ReadMarkers.argtypes = [ ctypes.POINTER(MarkerReader), ctypes.c_uint64, ctypes.POINTER(ctypes.c_uint64) ]
    c_ReadMarkers.restype = ErrorCode
    c_ReadMarkers.errcheck = HandleErrorCode
    recordsRead = ctypes.c_uint64()
    c_ReadMarkers(reader, recordsToRead, ctypes.byref(recordsRead))
    return recordsRead.value

def MarkerReader_SetCallbacks(reader, callbacks, userData):
    c_SetCallbacks = conf.lib.OTF2_MarkerReader_SetCallbacks
    c_SetCallbacks.argtypes = [ ctypes.POINTER(MarkerReader), ctypes.POINTER(MarkerReaderCallbacks), ctypes.py_object ]
    c_SetCallbacks.restype = ErrorCode
    c_SetCallbacks.errcheck = HandleErrorCode
    c_SetCallbacks(reader, callbacks, ctypes.py_object(userData))

__all__ = [
    'MarkerReader',
    'MarkerReader_ReadMarkers',
    'MarkerReader_SetCallbacks'
]



#!/usr/bin/env python

import os
import os.path
import unittest
from shutil import rmtree
from tempfile import mkdtemp

import _otf2


def pre_flush(user_data, file_type, location, caller_data, final):
    del file_type, location, caller_data, final
    return _otf2.FLUSH


def post_flush(user_data, file_type, location):
    del file_type, location
    return get_time()


def get_time():
    get_time.counter += 1
    return get_time.counter

get_time.counter = -1
read_locations = []

global_time = 0
global_region = 0


class NestedCallbacks(object):
    def __init__(self, outter):
        self.outter = outter

    def location_callback(self, userData, id, name, type, num_events, group):
        read_locations.append(id)
        self.outter.calls += 1


class MyCallbacks(object):

    def __init__(self):
        self.calls = 0
        self.events = 0
        self.nested = NestedCallbacks(self)

    def set_location_callback(self, cbs):
        _otf2.GlobalDefReaderCallbacks_SetLocationCallback(cbs, self.nested.location_callback)
        _otf2.GlobalDefReaderCallbacks_SetStringCallback(cbs, self.string_callback)

    def string_callback(self, *args):
        print(args)

    @classmethod
    def enter_callback(cls, location, time, userData, attributes, region):
        assert time >= global_time
        assert region == global_region

    @staticmethod
    def leave_callback(location, time, userData, attributes, region):
        global global_time, global_region
        assert time >= global_time
        global_time = time

        assert region == global_region
        global_region = (global_region + 1) % 2

    def mpi_send_callback(self, location, time, userData, attributes, receiver, comm, tag, length):
        global global_time
        assert time >= global_time
        global_time = time

        self.events += 1

        assert location == receiver
        assert comm == 0
        assert tag == 42
        assert length == 1024


def mpi_recv_callback(location, time, userData, attributes, sender, comm, tag, length):
    global global_time
    assert time >= global_time
    global_time = time

    assert location == sender
    assert comm == 0
    assert tag == 42
    assert length == 1024


class TestOTF2ReaderCallbacks(unittest.TestCase):

    num_locations = 4
    num_events = 200
    archive_name = "OTF2_Integrity_trace"
    locations = [x * x for x in range(num_locations)]

    def setUp(self):
        self.old_cwd = os.getcwd()
        self.tmp_dirname = mkdtemp(prefix=os.path.basename(os.path.abspath(__file__))[:-3] + '_tmp', dir=self.old_cwd)
        os.chdir(self.tmp_dirname)

    def tearDown(self):
        os.chdir(self.old_cwd)
        if os.getenv('KEEP_TEST_OUTPUT', '') != '':
            print(self.tmp_dirname)
        else:
            rmtree(self.tmp_dirname)

    def write_trace(self):
        archive_path = "%s/%s" % (self.tmp_dirname, self.archive_name)
        archive = _otf2.Archive_Open(archive_path,
                                     "trace_files",
                                     _otf2.FILEMODE_WRITE,
                                     256 * 1024,
                                     _otf2.UNDEFINED_UINT64,
                                     _otf2.SUBSTRATE_POSIX,
                                     _otf2.COMPRESSION_NONE)

        flush_callbacks = _otf2.FlushCallbacks(pre_flush=pre_flush,
                                               post_flush=post_flush)
        _otf2.Archive_SetFlushCallbacks(archive,
                                        flushCallbacks=flush_callbacks,
                                        flushData=None)

        _otf2.Archive_SetSerialCollectiveCallbacks(archive)

        _otf2.Archive_SetDescription(archive, description="OTF2 trace for integrity test.")
        _otf2.Archive_SetCreator(archive, creator="OTF2_Integrity")

        _otf2.Archive_SetDefChunkSize(archive, 4 * 1024 * 1024)

        global_def_writer = _otf2.Archive_GetGlobalDefWriter(archive)

        _otf2.GlobalDefWriter_WriteString(global_def_writer, 0, "Process")

        for location in self.locations:
            _otf2.GlobalDefWriter_WriteLocation(
                global_def_writer, location, 0, _otf2.LOCATION_TYPE_CPU_THREAD, self.num_events * 6, 0)

        _otf2.GlobalDefWriter_WriteString(global_def_writer, 1, "MPI_Send")
        _otf2.GlobalDefWriter_WriteString(global_def_writer, 2, "Send an MPI message")

        _otf2.GlobalDefWriter_WriteString(global_def_writer, 3, "MPI_Recv")
        _otf2.GlobalDefWriter_WriteString(global_def_writer, 4, "Receive an MPI message")

        _otf2.GlobalDefWriter_WriteRegion(
            global_def_writer, 0, 1, 1, 2, _otf2.REGION_ROLE_UNKNOWN, _otf2.PARADIGM_UNKNOWN, _otf2.REGION_FLAG_NONE, 0, 0, 0)
        _otf2.GlobalDefWriter_WriteRegion(
            global_def_writer, 1, 3, 3, 4, _otf2.REGION_ROLE_UNKNOWN, _otf2.PARADIGM_UNKNOWN, _otf2.REGION_FLAG_NONE, 0, 0, 0)

        _otf2.Archive_CloseGlobalDefWriter(archive, global_def_writer)

        _otf2.Archive_OpenDefFiles(archive)

        for location in self.locations:
            def_writer = _otf2.Archive_GetDefWriter(archive, location)
            _otf2.Archive_CloseDefWriter(archive, def_writer)

        _otf2.Archive_CloseDefFiles(archive)

        _otf2.Archive_OpenEvtFiles(archive)

        for _ in range(self.num_events):
            for location in self.locations:
                evt_writer = _otf2.Archive_GetEvtWriter(archive, location)

                ts = get_time()

                _otf2.EvtWriter_Enter(evt_writer, None, ts, 0)
                _otf2.EvtWriter_MpiSend(evt_writer, None, ts, location, 0, 42, 1024)
                _otf2.EvtWriter_Leave(evt_writer, None, get_time(), 0)

                _otf2.EvtWriter_Enter(evt_writer, None, get_time(), 1)
                ts = get_time()
                _otf2.EvtWriter_MpiRecv(evt_writer, None, ts, location, 0, 42, 1024)
                _otf2.EvtWriter_Leave(evt_writer, None, ts, 1)

        for location in self.locations:
            evt_writer = _otf2.Archive_GetEvtWriter(archive, location)
            _otf2.Archive_CloseEvtWriter(archive, evt_writer)

        _otf2.Archive_CloseEvtFiles(archive)

        _otf2.Archive_Close(archive)


    def read_trace(self):
        archive_path = "%s/%s/%s.otf2" % (self.tmp_dirname, self.archive_name, "trace_files")
        reader = _otf2.Reader_Open(archive_path)
        _otf2.Reader_SetSerialCollectiveCallbacks(reader)

        callbacks = MyCallbacks()

        num_locations = _otf2.Reader_GetNumberOfLocations(reader)
        assert num_locations == self.num_locations

        global_def_reader = _otf2.Reader_GetGlobalDefReader(reader)

        global_def_reader_callbacks = _otf2.GlobalDefReaderCallbacks_New()

        callbacks.set_location_callback(global_def_reader_callbacks)

        _otf2.Reader_RegisterGlobalDefCallbacks(reader, global_def_reader,
                                                global_def_reader_callbacks, None)

        _otf2.GlobalDefReaderCallbacks_Delete(global_def_reader_callbacks)

        definitions_read = _otf2.Reader_ReadGlobalDefinitions(
            reader, global_def_reader, _otf2.UNDEFINED_UINT64)

        _otf2.Reader_CloseGlobalDefReader(reader, global_def_reader)

        assert definitions_read == _otf2.Reader_GetNumberOfGlobalDefinitions(reader)
        assert callbacks.calls == self.num_locations

        for location in read_locations:
            _otf2.Reader_SelectLocation(reader, location)

        _otf2.Reader_OpenDefFiles(reader)
        _otf2.Reader_OpenEvtFiles(reader)

        for location in read_locations:
            def_reader = _otf2.Reader_GetDefReader(reader, location)
            definitions_read = _otf2.Reader_ReadAllLocalDefinitions(reader, def_reader)
            _otf2.Reader_CloseDefReader(reader, def_reader)
            evt_reader = _otf2.Reader_GetEvtReader(reader, location)

        _otf2.Reader_CloseDefFiles(reader)

        global_evt_reader = _otf2.Reader_GetGlobalEvtReader(reader)

        event_cbs = _otf2.GlobalEvtReaderCallbacks_New()
        _otf2.GlobalEvtReaderCallbacks_SetEnterCallback(event_cbs, MyCallbacks.enter_callback)
        _otf2.GlobalEvtReaderCallbacks_SetLeaveCallback(event_cbs, MyCallbacks.leave_callback)
        _otf2.GlobalEvtReaderCallbacks_SetMpiSendCallback(event_cbs, callbacks.mpi_send_callback)
        _otf2.GlobalEvtReaderCallbacks_SetMpiRecvCallback(event_cbs, mpi_recv_callback)
        _otf2.Reader_RegisterGlobalEvtCallbacks(reader, global_evt_reader, event_cbs, None)
        _otf2.GlobalEvtReaderCallbacks_Delete(event_cbs)

        events_read = _otf2.Reader_ReadGlobalEvents(
            reader, global_evt_reader, _otf2.UNDEFINED_UINT64)

        assert events_read == 6 * self.num_locations * self.num_events
        assert callbacks.events == self.num_events * self.num_locations

        _otf2.Reader_CloseGlobalEvtReader(reader, global_evt_reader)
        _otf2.Reader_CloseEvtFiles(reader)
        _otf2.Reader_Close(reader)

    def test_integrity(self):
        self.write_trace()
        self.read_trace()


if __name__ == '__main__':
    unittest.main()
